import unittest
import unittest.mock
import numpy

from osgar.lib.serialize import serialize, deserialize, compress


class SerializeTest(unittest.TestCase):

    def test_serialization(self):
        self.assertEqual(serialize(b'\x01\x02'), b'\xc4\x02\x01\x02')
        position = (51749517, 180462688)
        self.assertEqual(serialize(position), b'\x92\xce\x03\x15\xa2\x8d\xce\n\xc1\xa4`')
        self.assertEqual(serialize((123.4, 'Hi')), b'\x92\xcb@^\xd9\x99\x99\x99\x99\x9a\xa2Hi')

    def test_packed_data(self):
        data = [0]*1000
        packet = serialize(data)
        self.assertEqual(len(packet), 1003)
        self.assertEqual(deserialize(packet), data)

        compressed_packet = compress(packet)
        self.assertEqual(len(compressed_packet), 22)
        self.assertEqual(deserialize(compressed_packet), data)

    def test_numpy(self):
        data = numpy.asarray(range(1, 100, 2), dtype=numpy.uint16)
        packet = serialize(data)
        packet_zlib = compress(packet)
        self.assertEqual(len(packet), 231)
        self.assertEqual(len(packet_zlib), 168)
        dedata = deserialize(packet)
        self.assertTrue(numpy.array_equal(dedata, data))
        self.assertEqual(dedata.dtype, data.dtype)

        dedata_zlib = deserialize(packet_zlib)
        self.assertTrue(numpy.array_equal(dedata_zlib, data))
        self.assertEqual(dedata_zlib.dtype, data.dtype)

        with unittest.mock.patch('osgar.lib.serialize.numpy', new=False):
            with self.assertRaises(TypeError):
                serialize(data)
            with self.assertRaises(TypeError):
                deserialize(packet)

# vim: expandtab sw=4 ts=4



import unittest

from osgar.lib.lidar_pts import equal_scans


class LidarPtsTest(unittest.TestCase):

    def test_equal_scans(self):
        self.assertTrue(equal_scans([1, 2, 3], [1, 2, 3]))
        self.assertFalse(equal_scans([1, 2, 3], [1, 4, 3], tollerance=1))


# vim: expandtab sw=4 ts=4



import ast
import datetime
import logging
import os
import pathlib
import shutil
import tempfile
import time
import unittest

from unittest.mock import MagicMock

from threading import Thread

import osgar.logger

from osgar.zmqrouter import _Router, _Bus, record


class Noop:
    def __init__(self, config, bus):
        self.bus = bus
        self.bus.register("output")

    def start(self):
        pass

    def join(self, timeout=None):
        pass


class Publisher:
    def __init__(self, config, bus):
        self.bus = bus
        self.output_name = config["output"].split(":")[0] # drop any possible suffix
        self.bus.register(config["output"])

    def start(self):
        self.thread = Thread(target=self.run)
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def run(self):
        count = 10
        for i in range(count):
            dt = self.bus.publish(self.output_name, i)
            time.sleep(0.01)
            #print("  published", i, dt)


class Sleeper:
    def __init__(self, config, bus):
        self.bus = bus
        self.bus.register()

    def start(self):
        self.bus.sleep(0.001)

    def join(self, timeout=None):
        pass


class Listener:
    def __init__(self, config, bus):
        self.bus = bus
        self.bus.register()

    def start(self):
        self.thread = Thread(target=self.run)
        self.thread.start()

    def run(self):
        while True:
            self.bus.listen()

    def join(self, timeout=None):
        self.thread.join(timeout)


class PublisherListener:
    def __init__(self, config, bus):
        self.bus = bus
        self.output_name = config["output"].split(":")[0] # drop any possible suffix
        self.bus.register(config["output"])

    def start(self):
        self.thread = Thread(target=self.run)
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def run(self):
        count = 10
        for i in range(count):
            dt, channel, value = self.bus.listen()
            assert value == i
            time.sleep(0.15)
            dt = self.bus.publish(self.output_name, i)
            #print(self.bus.name, dt, channel, value)


class ThreadedPublisher:
    def __init__(self, config, bus):
        self.bus = bus
        self.bus.register('raw')
        self.input_thread = None
        self.output_thread = None

    def start(self):
        self.input_thread = Thread(target=self.run_input)
        self.output_thread = Thread(target=self.run_output)
        self.input_thread.start()
        self.output_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout)
        self.output_thread.join(timeout)

    def run_input(self):
        for i in range(10):
            self.bus.publish('raw', b'data from outside')
            if not self.bus.is_alive():
                return

    def run_output(self):
        while True:
            dt, channel, data = self.bus.listen()


class NoQuit:
    def __init__(self, config, bus):
        self.bus = bus
        self.bus.register()

    def start(self):
        self.thread = Thread(target=self.run)
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def run(self):
        while True:
            time.sleep(1)


class Test(unittest.TestCase):
    def setUp(self):
        self.tempdir = pathlib.Path(tempfile.mkdtemp(dir=pathlib.Path.cwd()))
        os.environ['OSGAR_LOGS'] = str(self.tempdir)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_threads(self):
        main()

    def test_noop(self):
        config = {
            'version': 2,
            'robot': {
                'modules': {
                    "noop": {
                        "driver": "osgar.test_zmqrouter:Noop",
                        "init": {}
                    },
                }, 'links':[]
            }
        }
        record(config, log_filename='noop.log')

    def test_publisher_single(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher'] = {
            "driver": "osgar.test_zmqrouter:Publisher",
            "init": { "output": "count"}
        }
        record(config, log_filename='publisher.log')
        with osgar.logger.LogReader(self.tempdir/"publisher.log", only_stream_id=1) as log:
            last_dt = datetime.timedelta()
            count = 0
            for dt, channel, data in log:
                self.assertGreater(dt, last_dt)
                self.assertEqual(int.from_bytes(data, 'little'), count)
                last_dt = dt
                count += 1

    def test_publisher_multi(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        for i in range(3):
            config['robot']['modules'][f'publisher{i}'] = {
                "driver": "osgar.test_zmqrouter:Publisher",
                "init": { "output": f"count{i}" }
            }
        record(config, log_filename='publisher.log')

    def test_publisher_threaded(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher-threaded'] = {
            "driver": "osgar.test_zmqrouter:ThreadedPublisher",
        }
        record(config, log_filename='publisher-threaded.log', duration_sec=3)
        with osgar.logger.LogReader(self.tempdir/"publisher-threaded.log", only_stream_id=1) as log:
            count = sum(1 for _ in log)
            self.assertEqual(count, 10)

    def test_compress(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher'] = {
            "driver": "osgar.test_zmqrouter:Publisher",
            "init": { "output": "count:gz"}
        }
        record(config, log_filename='compressed-publisher.log')

    def test_null(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher'] = {
            "driver": "osgar.test_zmqrouter:Publisher",
            "init": { "output": "count:null"}
        }
        record(config, log_filename='null-publisher.log')

    def test_delays(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher-listener'] = {
            "driver": "osgar.test_zmqrouter:PublisherListener",
            "init": { "output": "count"}
        }
        config['robot']['modules']['publisher'] = {
            "driver": "osgar.test_zmqrouter:Publisher",
            "init": { "output": "count"}
        }
        config['robot']['links'] = [
            ['publisher.count', 'publisher-listener.count'],
        ]
        record(config, log_filename='delays.log')
        with osgar.logger.LogReader(self.tempdir/"delays.log", only_stream_id=0) as log:
            count = 0
            for dt, channel, data in log:
                data = ast.literal_eval(str(data, 'ascii'))
                if hasattr(data, 'keys') and 'delay' in data.keys():
                    count += 1
                    self.assertGreater(data['delay'], 0.15)
            self.assertEqual(count, 10)

    def test_sleep(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['sleeper'] = {
            "driver": "osgar.test_zmqrouter:Sleeper",
        }
        record(config, log_filename='sleeps.log')

    def test_duration(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['listener'] = {
            "driver": "osgar.test_zmqrouter:Listener",
        }
        record(config, log_filename='duration.log', duration_sec=0.3)

    @unittest.skip("until we figure out what to do with the timeout")
    def test_fail_to_register(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher'] = {
            "driver": "osgar.test_zmqrouter:Nonexisting",
            "init": { "output": "count:null"}
        }
        record(config, log_filename='null-publisher.log')

    def test_fail_to_quit(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['publisher'] = {
            "driver": "osgar.test_zmqrouter:Publisher",
            "init": { "output": "count:null"}
        }
        config['robot']['modules']['noquit'] = {
            "driver": "osgar.test_zmqrouter:NoQuit",
        }
        record(config, log_filename='noquit.log')

    def test_fail_to_quit_duration(self):
        config = { 'version': 2, 'robot': { 'modules': {}, 'links': [] } }
        config['robot']['modules']['noquit'] = {
            "driver": "osgar.test_zmqrouter:NoQuit",
        }
        record(config, log_filename='duration-no-quit.log', duration_sec=0.3)


def main():
    nodes = ["listener0", "listener1", "publisher"]
    links = [
        ["publisher.count", "listener0.count"],
        ["publisher.count", "listener1.count"],
    ]

    logger = MagicMock()
    logger.start_time = datetime.datetime.now(datetime.timezone.utc)
    with _Router(logger) as router:
        Thread(target=node_listener, args=("listener0",)).start()
        Thread(target=node_listener, args=("listener1",)).start()
        Thread(target=node_publisher, args=("publisher", "count", 10)).start()

        router.register_nodes(nodes)
        for link_from, link_to in links:
            router.connect(link_from, link_to)
        router.run()
    #print(logger.mock_calls)


def node_listener(name):
    bus = _Bus(name)
    bus.register()
    try:
        expected = 0
        while True:
            dt, channel, data = bus.listen()
            assert data == expected
            expected += 1
            #print(f"  {name}", dt, channel, data)
    except SystemExit:
        bus.request_stop()


def node_publisher(name, channel, count):
    bus = _Bus(name)
    bus.register(channel)
    for i in range(count):
        dt = bus.publish(channel, i)
        #print("  published", i, dt)
    bus.request_stop()


if __name__ == "__main__":
    import sys
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.DEBUG,
        format='%(asctime)s %(name)-16s %(levelname)-8s %(message)s',
    )
    unittest.main()



import unittest
from unittest.mock import MagicMock
from datetime import timedelta

from osgar.drivers.imu import parse_line, IMU
from osgar.bus import Bus


class IMUTest(unittest.TestCase):

    nmea_line = b'$VNYMR,-095.878,+008.933,-009.419,+00.1785,+00.1474,+00.5479,+00.252,-03.706,-03.874,-01.012502,-00.234810,+00.247165*66'
    orientation = [[-95.878, 8.933, -9.419], [0.1785, 0.1474, 0.5479], [0.252, -3.706, -3.874], [-1.012502, -0.23481, 0.247165]]

    def test_parse_line(self):
        # (yaw, pitch, roll), (magx, y, z), (accx, y, z), (gyrox, y, z)
        self.assertEqual(len(parse_line(self.nmea_line)), 4)

    def test_start_stop(self):
        config = {}
        logger = MagicMock()
        bus = Bus(logger)
        imu = IMU(config, bus=bus.handle('imu'))
        imu.start()
        imu.request_stop()
        imu.join()

    def test_processing(self):
        config = {}
        logger = MagicMock()
        logger.write = MagicMock(return_value=timedelta(135))
        bus = Bus(logger)
        #robot_bus = BusHandler(logger, out={}, name='robot')
        #bus = BusHandler(logger,
        #                 out={'orientation':[(robot_bus.queue, 'orientation')], 'rotation':[]},
        #                 name='imu')
        imu = IMU(config, bus=bus.handle('imu'))
        tester = bus.handle('tester')
        tester.register('raw')
        bus.connect('tester.raw', 'imu.raw')
        bus.connect('imu.orientation', 'tester.orientation')
        imu.start()
        tester.publish('raw', self.nmea_line)
        imu.request_stop()
        imu.join()
        tester.shutdown()
        self.assertEqual(tester.listen(), (timedelta(135), 'orientation', self.orientation))

# vim: expandtab sw=4 ts=4



"""
  RoboOrienteering 2018 - experiments with osgar.logger and robot container
"""

import argparse
import math
import threading

from datetime import timedelta

from osgar.lib.config import config_load
from osgar.lib.mathex import normalizeAnglePIPI
from osgar.record import record

from osgar.drivers.gps import INVALID_COORDINATES


def geo_length(pos1, pos2):
    "return distance on sphere for two integer positions in milliseconds"
    x_scale = math.cos(math.radians(pos1[0]/3600000))
    scale = 40000000/(360*3600000)
    return math.hypot((pos2[0] - pos1[0])*x_scale, pos2[1] - pos1[1]) * scale


def geo_angle(pos1, pos2):
    if geo_length(pos1, pos2) < 1.0:
        return None
    x_scale = math.cos(math.radians(pos1[0]/3600000))
    return math.atan2(pos2[1] - pos1[1], (pos2[0] - pos1[0])*x_scale)


def latlon2xy(lat, lon):
    return int(round(lon*3600000)), int(round(lat*3600000))

class EmergencyStopException(Exception):
    pass

class EmergencyStopMonitor:
    def __init__(self, robot):
        self.robot = robot

    def update(self, robot):
        if (robot.status & RoboOrienteering2018.EMERGENCY_STOP) == 0:
            raise EmergencyStopException()

    # context manager functions
    def __enter__(self):
        self.callback = self.robot.register(self.update)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.robot.unregister(self.callback)


class RoboOrienteering2018:
    EMERGENCY_STOP = 0x0001

    def __init__(self, config, bus):
        self.bus = bus
        self.maxspeed = config['maxspeed']
        self.goals = [latlon2xy(lat, lon) for lat, lon in config['waypoints']]
        self.time = None
        self.last_position = None  # (lon, lat) in milliseconds
        self.last_imu_yaw = None  # magnetic north in degrees
        self.status = None
        self.wheel_heading = None
        self.cmd = [0, 0]
        self.monitors = []
        self.last_position_angle = None  # for angle computation from dGPS

    def update(self):
        packet = self.bus.listen()
        if packet is not None:
#            print('RO', packet)
            timestamp, channel, data = packet
            self.time = timestamp
            if channel == 'position':
                self.last_position = data
            elif channel == 'orientation':
                (yaw, pitch, roll), (magx, y, z), (accx, y, z), (gyrox, y, z) = data
                self.last_imu_yaw = yaw
            elif channel == 'status':  # i.e. I can drive only spider??
                self.status, steering_status = data
                if steering_status is None:
                    self.wheel_heading = None
                else:
                    self.wheel_heading = math.radians(-360 * steering_status[0] / 512)
                self.bus.publish('move', self.cmd)
            for monitor_update in self.monitors:
                monitor_update(self)

    def set_speed(self, desired_speed, desired_wheel_heading):
        # TODO split this for Car and Spider mode
        speed = int(round(desired_speed))
        desired_steering = int(-512 * math.degrees(desired_wheel_heading) / 360.0)

        if speed != 0:
            if self.wheel_heading is None:
                speed = 1  # in in place
            else:
                 d = math.degrees(normalizeAnglePIPI(self.wheel_heading - desired_wheel_heading))
                 if abs(d) > 20.0:
                     speed = 1  # turn in place (II.)

        self.cmd = [speed, desired_steering]

    def start(self):
        self.thread = threading.Thread(target=self.play)
        self.thread.start()

    def request_stop(self):
        self.bus.shutdown()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def register(self, callback):
        self.monitors.append(callback)
        return callback

    def unregister(self, callback):
        assert callback in self.monitors
        self.monitors.remove(callback)

    def wait(self, dt):
        if self.time is None:
            self.update()
        start_time = self.time
        while self.time - start_time < dt:
            self.update()

    def play0(self):
        self.wait(timedelta(seconds=1))
        self.set_speed(1, math.radians(50))
        self.wait(timedelta(seconds=5))
        self.set_speed(10, 0)
        self.wait(timedelta(seconds=5))
        self.set_speed(0, 0)
        self.wait(timedelta(seconds=1))

    def play(self):
        print("Waiting for valid GPS position...")
        while self.last_position is None or self.last_position == INVALID_COORDINATES:
            self.update()
        print(self.last_position)

        print("Wait for valid IMU...")
        while self.last_imu_yaw is None:
            self.update()
        print(self.last_imu_yaw)

        print("Wait for steering info...")
        while self.wheel_heading is None:
            self.update()
        print(math.degrees(self.wheel_heading))

        print("Ready", self.goals)
        try:
            with EmergencyStopMonitor(self):
                for goal in self.goals:
                    print("Goal at %.2fm" % geo_length(self.last_position, goal))
                    angle = geo_angle(self.last_position, goal)
                    if angle is not None:
                        print("Heading %.1fdeg, imu" % math.degrees(angle), self.last_imu_yaw)
                    else:
                        print("Heading None, imu", self.last_imu_yaw)
                    self.navigate_to_goal(goal, timedelta(seconds=200))
        except EmergencyStopException:
            print("EMERGENCY STOP (wait 3s)")
            self.set_speed(0, 0)
            start_time = self.time
            while self.time - start_time < timedelta(seconds=3):
                self.set_speed(0, 0)
                self.update()

    def navigate_to_goal(self, goal, timeout):
        start_time = self.time
        self.last_position_angle = self.last_position
        gps_angle = None
        while geo_length(self.last_position, goal) > 1.0 and self.time - start_time < timeout:
            desired_heading = normalizeAnglePIPI(geo_angle(self.last_position, goal))
            step = geo_length(self.last_position, self.last_position_angle)
            if step > 1.0:
                gps_angle = normalizeAnglePIPI(geo_angle(self.last_position_angle, self.last_position))
                print('step', step, math.degrees(gps_angle))
                self.last_position_angle = self.last_position
                desired_wheel_heading = normalizeAnglePIPI(desired_heading - gps_angle + self.wheel_heading)

            if gps_angle is None or self.wheel_heading is None:
                spider_heading = normalizeAnglePIPI(math.radians(180 - self.last_imu_yaw - 35.5))
                desired_wheel_heading = normalizeAnglePIPI(desired_heading-spider_heading)

            self.set_speed(self.maxspeed, desired_wheel_heading)

            prev_time = self.time
            self.update()

            if int(prev_time.total_seconds()) != int(self.time.total_seconds()):
                print(self.time, geo_length(self.last_position, goal), self.last_imu_yaw, self.wheel_heading)

        print("STOP (3s)")
        self.set_speed(0, 0)
        start_time = self.time
        while self.time - start_time < timedelta(seconds=3):
            self.set_speed(0, 0)
            self.update()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='RoboOrienteering 2018')
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')
    subparsers.required = True
    parser_run = subparsers.add_parser('run', help='run on real HW')
    parser_run.add_argument('config', nargs='+', help='configuration file')
    parser_run.add_argument('--note', help='add description')

    parser_replay = subparsers.add_parser('replay', help='replay from logfile')
    parser_replay.add_argument('logfile', help='recorded log file')
    parser_replay.add_argument('--force', '-F', dest='force', action='store_true', help='force replay even for failing output asserts')
    parser_replay.add_argument('--config', nargs='+', help='force alternative configuration file')
    args = parser.parse_args()

    if args.command == 'replay':
        from osgar.replay import replay
        args.module = 'app'
        app = replay(args, application=RoboOrienteering2018)
        app.play()

    elif args.command == 'run':
        cfg = config_load(*args.config, application=RoboOrienteering2018)
        record(cfg, 'ro2018-')
    else:
        assert False, args.command  # unsupported command

# vim: expandtab sw=4 ts=4



import unittest

import numpy as np

# if this variable is set, stack trace on test failure from this module is not in the output
__unittest = True

class TestCase(unittest.TestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.addTypeEqualityFunc(np.ndarray, 'assertNumpyEqual')

    def assertNumpyEqual(self, n1, n2, msg=None):
        self.assertIsInstance(n1, np.ndarray, 'First argument is not a numpy.ndarray')
        self.assertIsInstance(n2, np.ndarray, 'Second argument is not a numpy.ndarray')
        np.testing.assert_almost_equal(n1, n2, err_msg=msg)

    def assertPose3dEqual(self, p1, p2):
        self.assertEqual(len(p1), 2, msg='First Pose3d needs xyz and quaternion')
        self.assertEqual(len(p2), 2, msg='Second Pose3d needs xyz and quaternion')
        self.assertEqual(len(p1[0]), 3, msg='First pose3d xyz needs 3 coordinates')
        self.assertEqual(len(p1[1]), 4, msg='First pose3d quaternion needs 4 coordinates')
        self.assertEqual(len(p2[0]), 3, msg='Second pose3d xyz needs 3 coordinates')
        self.assertEqual(len(p2[1]), 4, msg='Second pose3d quaternion needs 4 coordinates')
        self.assertListAlmostEqual(p1[0], p2[0])
        self.assertListAlmostEqual(p1[1], p2[1])

    def assertListAlmostEqual(self, l1, l2, places=None, delta=None):
        self.assertIsInstance(l1, list, 'First argument is not a list')
        self.assertIsInstance(l2, list, 'Second argument is not a list')
        self.assertEqual(len(l1), len(l2), "")
        for i, (a, b) in enumerate(zip(l1, l2)):
            self.assertAlmostEqual(a, b, places=places, msg=f"\nFirst differing element index: {i}", delta=delta)



import unittest

from subt.name_decoder import parse_robot_name, split_multi


class NameDecoderTest(unittest.TestCase):
    def test_parse_robot_name(self):
        steps = parse_robot_name('A10W100L')
        self.assertEqual(steps, [('wait', 10), ('enter-left', None), ('left', 100), ('home', 200)])

    def test_both_sides(self):
        steps = parse_robot_name('A10W100LH200R')
        self.assertEqual(steps, [('wait', 10), ('enter-left', None), ('left', 100), ('home', 200), ('enter-right', None), ('right', 200), ('home', 400)])

    def test_variable_side(self):
        steps = parse_robot_name('A10W100L200R')
        self.assertEqual(steps, [('wait', 10), ('enter-left', None), ('left', 100), ('right', 200), ('home', 600)])

    def test_no_wait(self):
        steps = parse_robot_name('A100L')
        self.assertEqual(steps, [('enter-left', None), ('left', 100), ('home', 200)])

    def test_split_multi(self):
        self.assertEqual(split_multi('123L456R78', ['L', 'R']), ['123L', '456R', '78'])

    def test_old_code(self):
        with self.assertRaises(ValueError):
            parse_robot_name('A10F100L')

    def test_double_home(self):
        steps = parse_robot_name('A100LH')
        self.assertEqual(steps, [('enter-left', None), ('left', 100), ('home', 200)])

    def test_home_side(self):
        # it turned out that "enter" is ambiguous as there is offset navigating the robot
        # closer to the wall needed for the next left/right/center command.
        # In order to simplify processing the type has to be integrated here.
        steps = parse_robot_name('A100L')
        self.assertEqual(steps, [('enter-left', None), ('left', 100), ('home', 200)])

        steps = parse_robot_name('A100C')
        self.assertEqual(steps, [('enter-center', None), ('center', 100), ('home', 200)])

# vim: expandtab sw=4 ts=4




"""
    Analyze 2D lidar scan points
"""
import math

import cv2
import numpy as np

from osgar.lib.line import distance


def rect(scan, debug_poly=None):
    """
    Dummy function to test debug polygons of lidar viewer
    """
    if debug_poly is not None:
        debug_poly.append([(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)])


def convex_hull(scan, debug_poly=None):
    """
    Test visualization of convex hull
    """
    heading = 0.0  # TODO
    pts = []
    for i, i_dist in enumerate(scan):
        if i_dist == 0 or i_dist >= 10000:
            continue
        angle = math.radians(270 * (i / len(scan)) - 135) + heading
        dist = i_dist/1000.0
        x, y = dist * math.cos(angle), dist * math.sin(angle)
        pts.append((x,y))

    hull = cv2.convexHull(np.array(pts, dtype=np.float32))
    hull_pts = [p[0] for p in hull] + [hull[0][0],]
    dist_arr = [distance(a, b) for a, b in zip(hull_pts[:-1], hull_pts[1:])]
    #print('%.2f %.2f %.2f %.2f' % tuple(sorted(dist_arr, reverse=True)[:4]))
    first, second = sorted(dist_arr, reverse=True)[:2]
    i, j = dist_arr.index(first), dist_arr.index(second)  # TODO argmax
    A = ((hull_pts[i][0] + hull_pts[i+1][0])/2, (hull_pts[i][1] + hull_pts[i+1][1])/2)
    B = ((hull_pts[j][0] + hull_pts[j+1][0])/2, (hull_pts[j][1] + hull_pts[j+1][1])/2)
    if debug_poly is not None:
        #debug_poly.append(hull_pts)
        debug_poly.append([A, B])
    return (A, B)


def dual_convex_hull(scan, debug_poly=None):
    """
    Compute convex hull for left and right side of the scan
    (usefull if already in tunnel)
    """
    heading = 0.0  # TODO
    pts_left, pts_right = [], []
    for i, i_dist in enumerate(scan):
        if i_dist == 0 or i_dist >= 10000:
            continue
        angle = math.radians(270 * (i / len(scan)) - 135) + heading
        dist = i_dist/1000.0
        x, y = dist * math.cos(angle), dist * math.sin(angle)
        if angle >= 0:  # beware of heading modification, TODO autodetect
            pts_left.append((x, y))
        else:
            pts_right.append((x, y))

    hull_left = cv2.convexHull(np.array(pts_left, dtype=np.float32))
    hull_right = cv2.convexHull(np.array(pts_right, dtype=np.float32))
    if debug_poly is not None:
        debug_poly.append([p[0] for p in hull_left])
        debug_poly[-1].append(hull_left[0][0])  # close polygon
        debug_poly.append([p[0] for p in hull_right])
        debug_poly[-1].append(hull_right[0][0])  # close polygon


def equal_scans(scan1, scan2, tollerance=10):
    """
    Compare two scans for "almost identity". Expected inputs are raw scans (in millimeters).
    """
    assert len(scan1) == len(scan2), (len(scan1), len(scan2))
    arr1 = np.array(scan1)
    arr2 = np.array(scan2)
    diff = arr1 - arr2
    return np.abs(diff).max() <= tollerance

# vim: expandtab sw=4 ts=4



from xml.sax.handler import ContentHandler
from xml.sax import make_parser

import base64
import zlib


class GazeboHandler(ContentHandler):
    def __init__(self):
        super().__init__()
        self.active = False
        self.index = 0

    def startElement(self, name, attrs):
        assert name in ['gazebo_log', 'header', 'log_version', 'gazebo_version',
                'rand_seed', 'chunk', 'log_start', 'log_end'], name

        if name == 'chunk':
            assert attrs['encoding'] == 'zlib', attrs['encoding']
            self.data = ''
        self.active = (name == 'chunk')

    def characters(self, data):
        if self.active:
            self.data += data

    def endElement(self, name):
        if name == 'chunk':
            subtree = zlib.decompress(base64.b64decode(self.data))
            if self.index == 100:
                print(subtree.decode('ascii'))
            self.index += 1


if __name__ == "__main__":
    import sys
    handler = GazeboHandler()

    saxparser = make_parser()
    saxparser.setContentHandler(handler)

    with open(sys.argv[1]) as f:
        saxparser.parse(f)


# vim: expandtab sw=4 ts=4



"""
  ROS (Robot Operating System) Message Parser
"""

from threading import Thread
import struct
import math

import numpy as np

from osgar.bus import BusShutdownException
from osgar.lib.quaternion import euler_zyx

from subt import ign_pb2

MAX_TOPIC_NAME_LENGTH = 256


ROS_MESSAGE_TYPES = {
    'std_msgs/String': '992ce8a1687cec8c8bd883ec73ca41d1',
    'geometry_msgs/Twist': '9f195f881246fdfa2798d1d3eebca84a',
    'std_msgs/Imu': '6a62c6daae103f4ff57a132d6f95cec2',
}

def prefix4BytesLen(s):
    "adding ROS length"
    if type(s) == str:
        s = bytes(s, encoding='ascii')
    return struct.pack("I", len(s)) + s


def packCmdVel(speed, angularSpeed):
    return struct.pack("dddddd", speed, 0, 0, 0, 0, angularSpeed)


def parse_imu( data ):
    # http://docs.ros.org/api/sensor_msgs/html/msg/Imu.html
    size = struct.unpack_from('<I', data)[0]
    #assert size in [324, 323], size  # expected IMU packet size (beware of variable frameIdLen - should be removed!)
    data = data[4:]
    seq, stamp, stampNsec, frameIdLen = struct.unpack("IIII", data[:16])
#    print(seq, stamp, stampNsec, frameIdLen)
#    print(data[16:16+frameIdLen])
    data = data[16+frameIdLen:]
    # variable frameIdLen:
    #    b'X1/base_link' = 12 bytes
    #    b'X3/imu_link' = 11 bytes
    assert len(data) == 296, len(data)  # fixed data size (37 * 8)
    orientation = struct.unpack("dddd", data[:4*8])
    data = data[4*8+9*8:] # skip covariance matrix
    angularVelocity = struct.unpack("ddd", data[:3*8])
    data = data[3*8+9*8:] # skip velocity covariance
    linearAcceleration = struct.unpack("ddd", data[:3*8])
#    print('%d\t%f' % (stamp, sum([x*x for x in linearAcceleration])))
    data = data[3*8+9*8:] # skip velocity covariance
    assert len(data) == 0, len(data)

    az, ay, ax = euler_zyx(orientation)  # quaternion, rotations along axis: yaw, pitch, roll

    return linearAcceleration, (az, ay, ax), orientation


def Xparse_image( data ):
    seq, stamp, stampNsec, frameIdLen = struct.unpack("IIII", data[:16])
#    print(seq, stamp, stampNsec, frameIdLen)
    data = data[16+frameIdLen:]

# from rosbag.py
def parse_raw_image(data, dump_filename=None):
    # http://docs.ros.org/melodic/api/sensor_msgs/html/msg/Image.html
    size = struct.unpack_from('<I', data)[0]
#    assert size == 230467, size  # expected size for raw image during experiment
    # http://docs.ros.org/melodic/api/std_msgs/html/msg/Header.html
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
#    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    height, width, encoding_size = struct.unpack_from('<III', data, pos)
    pos += 4 + 4 + 4
    encoding = data[pos:pos+encoding_size]
    pos += encoding_size
#    print(height, width, encoding)
    is_bigendian, step, image_arr_size = struct.unpack_from('<BII', data, pos)
    pos += 1 + 4 + 4

    if encoding == b'32FC1':
        # depth is array of floats, OSGAR uses uint16 in millimeters
        # cut min & max (-inf and inf are used for clipping)
        arr = np.frombuffer(data[pos:pos + image_arr_size], dtype=np.dtype('f'))*1000
        arr = np.clip(arr, 1, 0xFFFF)
        arr = np.ndarray.astype(arr, dtype=np.dtype('H'))
    elif encoding == b'16UC1':
        # depth is array as uint16, similar to OSGAR
        arr = np.frombuffer(data[pos:pos + image_arr_size], dtype=np.dtype('H'))

    elif encoding == b'rgb8':
        # not compressed image data
        img = np.frombuffer(data[pos:pos+image_arr_size], dtype=np.dtype('B'))
        img = np.reshape(img, (height, width, 3))
        return img

    else:
        assert False, encoding  # unsuported encoding

    if dump_filename is not None:
        with open(dump_filename, 'wb') as f:
            # RGB color format (PPM - Portable PixMap)
#            f.write(b'P6\n%d %d\n255\n' % (width, height))
#            f.write(data[pos:pos+image_arr_size])
            # Grayscale float format (PGM - Portable GrayMap)
            f.write(b'P5\n%d %d\n255\n' % (width, height))
            f.write(bytes([min(255, x//100) for x in arr]))
    return np.array(arr).reshape((height, width))


def parse_jpeg_image(data, dump_filename=None):
    # http://docs.ros.org/api/sensor_msgs/html/msg/CompressedImage.html
    size = struct.unpack_from('<I', data)[0]
    # http://docs.ros.org/melodic/api/std_msgs/html/msg/Header.html
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
#    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    encoding_size = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    encoding = data[pos:pos+encoding_size]
    pos += encoding_size
#    print(encoding, size)
    img_size = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    if dump_filename is not None:
        with open(dump_filename, 'wb') as f:
            f.write(data[pos:])
    return data[pos:]


def parse_laser(data):
    # http://docs.ros.org/melodic/api/sensor_msgs/html/msg/LaserScan.html
    size = struct.unpack_from('<I', data)[0]
    #assert size == 5826, size  # expected size for short lidar
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
#    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    params = struct.unpack_from('<fffffff', data, pos)
    # angle_min, angle_max, angle_increment, time_increment, scan_time
    # range_min, range_max
#    print(params)
    angle_range_deg = int(round(math.degrees(params[1] - params[0])))
    pos += 7*4
    ranges_size = struct.unpack_from('<I', data, pos)[0]
    # assert ranges_size == 720, ranges_size
    pos += 4
    scan = struct.unpack_from('<' + 'f' * ranges_size, data, pos)
    pos += 4 * ranges_size
    intensities_size = struct.unpack_from('<I', data, pos)[0]
    # assert intensities_size == 720, intensities_size
    intensities = struct.unpack_from('<' + 'f' * intensities_size, data, pos)

    # conversion to int and millimeters
    scan = [int(x*1000) if x < 65.0 else 0 for x in scan]

    if angle_range_deg == 270:
        # SubT Virtual World
        return scan

    if angle_range_deg == 149:
        # NASA moon
        size = len(scan)
        assert size == 100, size  # by single sample
        ret = [0]*40 + scan + [0]*40
        return ret

    # scan from map for MOBoS, Maria, K2, ...
    to_cut = int(len(scan) / 360 * 45)  # original scan is 360deg
    scan = scan[to_cut:len(scan)-to_cut]
    return scan


def parse_posestamped(data):
    size = struct.unpack_from('<I', data)[0]
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
#    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/PoseStamped.html
    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/Quaternion.html

    x, y, z = struct.unpack_from('<ddd', data, pos)
    pos += 3 * 8
    ori = struct.unpack_from('<dddd', data, pos)
    pos += 4 * 8

    #q0, q1, q2, q3 = ori  # quaternion
    #ax =  math.atan2(2*(q0*q1+q2*q3), 1-2*(q1*q1+q2*q2))
    #ay =  math.asin(2*(q0*q2-q3*q1))
    #az =  math.atan2(2*(q0*q3+q1*q2), 1-2*(q2*q2+q3*q3))
    return (x, y, z), (ori)

def parse_odom(data):
    # http://docs.ros.org/melodic/api/nav_msgs/html/msg/Odometry.html
    size = struct.unpack_from('<I', data)[0]
    #assert size in [719, 724], size  # expected size for odometry (beware of variable header! i.e. this assert is wrong in general)
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
#    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    child_frame_id = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    #print(data[pos:pos + child_frame_id])
    pos += child_frame_id
    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/PoseWithCovariance.html
    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/Pose.html
    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/Point.html
    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/Quaternion.html
    x, y, z = struct.unpack_from('<ddd', data, pos)
    pos += 3 * 8
    ori = struct.unpack_from('<dddd', data, pos)
    pos += 4 * 8
    # float64[36] covariance
    pos += 36 * 8
    #print(x, y, z)

    # http://docs.ros.org/melodic/api/geometry_msgs/html/msg/Twist.html
    linear = struct.unpack_from('<ddd', data, pos)
    pos += 3 * 8
    angular = struct.unpack_from('<ddd', data, pos)
    pos += 3 * 8
    # float64[36] covariance
    pos += 36 * 8
    assert pos == len(data), (pos, len(data))

    #q0, q1, q2, q3 = ori  # quaternion
    #ax =  math.atan2(2*(q0*q1+q2*q3), 1-2*(q1*q1+q2*q2))
    #ay =  math.asin(2*(q0*q2-q3*q1))
    #az =  math.atan2(2*(q0*q3+q1*q2), 1-2*(q2*q2+q3*q3))
    return timestamp_sec, (x, y), euler_zyx(ori)


def parse_points(data):
    # http://docs.ros.org/melodic/api/sensor_msgs/html/msg/PointCloud2.html
    size = struct.unpack_from('<I', data)[0]
#    assert size == 2457728, size  # expected size for cloud points (beware of variable header! i.e. this assert is wrong in general)
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
#    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    height, width = struct.unpack_from('<II', data, pos)
    pos += 8
#    print(height, width)
    point_field_size = struct.unpack_from('<I', data, pos)[0]
    pos += 4
#     assert point_field_size == 4, point_field_size  # RGBD? - cloud == 5
#    print(data[pos:pos + 20])
    for i in range(point_field_size):
        str_size = struct.unpack_from('<I', data, pos)[0]
        pos += 4
        assert str_size < 10, str_size
        print(data[pos:pos + str_size])
        pos += str_size
        offset, datatype, count = struct.unpack_from('<IBI', data, pos)
        print(offset, datatype, count)
        assert datatype in [7, 4], datatype  # uint8 FLOAT32 = 7, uint8 UINT16=4
        assert count == 1, count
        pos += 9

    is_bigendian, point_step, row_step = struct.unpack_from('<?II', data, pos)
    pos += 9
#    print(is_bigendian, point_step, row_step)
    arr_size = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    assert arr_size == height * width * 32, arr_size
    for i in range(height*width):
        pt = struct.unpack_from('<ffff', data, pos + i * 32)
        assert pt[3] == 0.0, pt
#        if not math.isnan(pt[0]):
#            print(pt[:3])
#        assert str(pt) == str((nan, nan, nan, 0.0)), (i, pt)

    pos += row_step * height
    is_dense = struct.unpack_from('<?', data, pos)[0]
    pos += 1
#    print(is_dense)
    assert pos == len(data), (pos, len(data))


def parse_clock(data):
    size = struct.unpack_from('<I', data)[0]
    assert size == 8, size  # clock does not contain type, only NTP time
    pos = 4
    timestamp_sec, timestamp_nsec = struct.unpack_from('<II', data, pos)
    return (timestamp_sec, timestamp_nsec)


def parse_bool(data):
    size = struct.unpack_from('<I', data)[0]
    assert size == 1, size  # std_msgs::Bool does not contain type, only 1 byte of data
    val = data[4]
    assert val in [0, 1], val
    return val != 0


def parse_joint_state(data):
    # sensor_msgs/JointState.msg
    # Header
    # string[] name
    # float64[] position
    # float64[] velocity
    # float64[] effort
    size = struct.unpack_from('<I', data)[0]
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
    pos += frame_id_size
    size = struct.unpack_from('<I', data, pos)[0]  # number of joints
    pos += 4
    name = []
    for i in range(size):
        str_len = struct.unpack_from('<I', data, pos)[0]
        pos += 4
        name.append(data[pos:pos+str_len])
        pos += str_len

    tmp = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    assert size == tmp, (size, tmp)  # number of joints
    position = struct.unpack_from('<' + 'd'*size, data, pos)
    pos += 8*size

    tmp = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    assert size == tmp, (size, tmp)  # number of joints
    velocity = struct.unpack_from('<' + 'd'*size, data, pos)
    pos += 8*size

    tmp = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    assert size == tmp, (size, tmp)  # number of joints
    effort = struct.unpack_from('<' + 'd'*size, data, pos)
    pos += 8*size

    return name, position, velocity, effort


def get_frame_id(data):
    size = struct.unpack_from('<I', data)[0]
    pos = 4
    if size == 8:
        # exception for /clock ?!
        return b'/clock'
    if size == 1:
        # exception for Bool -> /gas_detected ?!
        return b'/gas_detected'

    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
    return frame_id


def parse_volatile(data):
    # NASA Space Robotics Challenge 2
    # __slots__ = ['header','vol_type','vol_index','shadowed_state','distance_to']
    # _slot_types = ['std_msgs/Header','string','int32','bool','float32']
    size = struct.unpack_from('<I', data)[0]
    pos = 4
    assert size + 4 == len(data), (size, len(data))  # it is going to be variable -> remove the assert
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]  # b'scout_1/volatile_sensor'
    pos += frame_id_size
    size = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    vol_type = data[pos:pos+size]  # b'methanol'
    pos += size
    vol_index, shadowed_state, distance_to = struct.unpack_from('<iBf', data, pos)
    return [vol_type.decode('ascii'), distance_to, vol_index]

def parse_bucket(data):
    # NASA Space Robotics Challenge 2
    # __slots__ = ['vol_type','vol_index','mass_in_bucket']
    # _slot_types = ['string','int32','float32']
    size = struct.unpack_from('<I', data)[0]
    pos = 4
    assert size + 4 == len(data), (size, len(data))  # it is going to be variable -> remove the assert
    vol_type_size = struct.unpack_from('<I', data, pos)[0]
    pos += 4
    vol_type = data[pos:pos+vol_type_size]  # b'methanol'
    pos += vol_type_size
    vol_index, mass_in_bucket = struct.unpack_from('<if', data, pos)
    return [vol_type.decode('ascii'), vol_index, mass_in_bucket]

def parse_topic(topic_type, data):
    """parse general topic"""
    if topic_type == 'srcp2_msgs/Qual1ScoringMsg':
        assert len(data) == 8, (len(data), data)
        size = struct.unpack_from('<I', data)[0]
        pos = 4
        assert size == 4, size
        # __slots__ = ['score']
        # _slot_types = ['int32']
        return struct.unpack_from('<I', data, pos)
    elif topic_type == 'srcp2_msgs/HaulerMsg':
        assert len(data) == 134, (len(data), data)
        # vol_type: [ice, ethene, methane, carbon_mono, carbon_dio, ammonia, hydrogen_sul, sulfur_dio]
        # mass_per_type: [0.0, 0.0, 0.0, 0.0, 0.0, 70.26217651367188, 0.0, 0.0]
        vol_type_arr = []
        mass_arr = []
        size = struct.unpack_from('<I', data)[0]
        pos = 4
        for i in range(8):
            vol_type_len = struct.unpack_from('<I', data, pos)[0]
            pos += 4
            vol_type = data[pos:pos+vol_type_len]
            vol_type_arr.append(vol_type.decode('ascii'))
            pos += vol_type_len
        for i in range(8):
            mass = struct.unpack_from('<f', data, pos)[0]
            mass_arr.append(mass)
            pos += 4
        return [vol_type_arr, mass_arr]
    elif topic_type == 'srcp2_msgs/Qual2ScoringMsg':
        assert len(data) == 142, (len(data), data)
        # __slots__ = ['vol_type', 'points_per_type', 'num_of_dumps', 'total_score']
        # _slot_types = ['string[8]', 'int32[8]', 'int32', 'float32']
        # let's ignore names of volatile types
        record = struct.unpack_from('<IIIIIIIIIf', data, len(data) - 10*4)
        # print(record)
        return [sum(record[:8]), record[8]]  # score and attempts
    elif topic_type == 'srcp2_msgs/Qual3ScoringMsg':
        assert len(data) == 12, (len(data), data)
        size = struct.unpack_from('<I', data)[0]
        pos = 4
        assert size == 8, size
        # __slots__ = ['score','calls']
        # _slot_types = ['int32','int32']
        return list(struct.unpack_from('<II', data, pos))  # score and calls
    elif topic_type == 'std_msgs/String':
        size = struct.unpack_from('<I', data)[0]
        pos = 4
        msg_len = struct.unpack_from('<I', data, pos)[0]
        pos += 4
        string = data[pos:pos+msg_len]
        return string.decode('ascii')
    elif topic_type == 'srcp2_msgs/ExcavatorMsg':
        return parse_bucket(data)
    elif topic_type == 'sensor_msgs/JointState':
        return parse_joint_state(data)
    elif topic_type == 'srcp2_msgs/VolSensorMsg':
        return parse_volatile(data)
    elif topic_type == 'sensor_msgs/LaserScan':
        return parse_laser(data)
    elif topic_type == 'geometry_msgs/PoseStamped':
        return parse_posestamped(data)
    elif topic_type == 'sensor_msgs/Imu':
        return parse_imu(data)
    elif topic_type == 'std_msgs/Bool':
        return parse_bool(data)
    elif topic_type == 'sensor_msgs/CompressedImage':
        image = parse_jpeg_image(data)  # , dump_filename='nasa.jpg')
        return image
    else:
        assert False, topic_type


class ROSMsgParser(Thread):
    def __init__(self, config, bus):
        Thread.__init__(self)
        self.setDaemon(True)
        outputs = ["rot", "acc", "scan", "image", "pose2d", "sim_time_sec", "sim_clock", "cmd", "origin", "gas_detected",
                   "depth:null", "t265_rot", "orientation", "debug", "radio", "base_station",
                    "joint_name", "joint_position", "joint_velocity", "joint_effort"]
        self.topics = config.get('topics', [])
        for row in self.topics:
            topic_name, topic_type = row[:2]
            if len(row) > 2:  # extra rename
                outputs.append(row[2])
            else:
                outputs.append(topic_name)  # default use topic name
        bus.register(*outputs)

        self.bus = bus
        self._buf = b''

        self.topic_type = config.get('topic_type')
        self.timestamp_sec = None
        self.timestamp_nsec = None

        # initial message contains structure
        self.header = None
        self.count = 0
        self.downsample = config.get('downsample', 1)

        self.desired_speed = None  # m/s
        self.desired_angular_speed = None
        # alternative 3D speed description used in ROS Twist message
        self.desired_speed_3d, self.desired_angular_speed_3d = None, None
        self.gas_detected = None  # this SubT Virtual specific :-(

        self.joint_name = None  # unknown

    def publish_desired_speed(self):
        if self.desired_speed is not None:
            cmd = b'cmd_vel %f %f' % (self.desired_speed, self.desired_angular_speed)
        elif self.desired_speed_3d is not None:
            cmd = b'cmd_vel_3d %f %f %f %f %f %f' % tuple(self.desired_speed_3d + self.desired_angular_speed_3d)
        else:
            # desired speed is not defined - valid state for waiting drone - do not send anything!
            return
        self.bus.publish('cmd', cmd)

    def get_packet(self):
        data = self._buf
        if len(data) < 4:
            return None
        size = struct.unpack_from('<I', data, 0)[0]
        print(size, len(self._buf))
        assert size > 0, size
        size += 4  # the length prefix
        if len(data) < size:
            return None
        ret, self._buf = data[:size], data[size:]
        return ret

    def slot_raw(self, timestamp, data):
#        self._buf += data
#        packet = self.get_packet()
#        if packet is None:
#            return
#        if self.header is None:
#            self.header = packet
#            return
#        self.count += 1
#        if self.count % self.downsample != 0:
#            return
        packet = data  # ZMQ hack
        if data.startswith(b'origin'):
            s = data.split()
            if b'ERROR' in s:
                self.bus.publish('origin', s[1:])
            else:
                self.bus.publish('origin', [s[1]] + [float(x) for x in s[2:]])
            return
        if data.startswith(b'depth'):
            depth = parse_raw_image(data[5:])
            self.bus.publish('depth', depth)
            return
        if data.startswith(b'radio '):
            s = data[6:].split(b' ')
            addr = s[0]
            msg = b' '.join(s[1:])
            if addr == b'base_station':
                artifact_score = ign_pb2.ArtifactScore()
                try:
                    artifact_score.ParseFromString(msg)
                except Exception as e:
                    print(e)
                    print(msg)
                    self.bus.report_error(exception=str(e), msg=msg)
                    return
                apos = [artifact_score.artifact.pose.position.x,
                        artifact_score.artifact.pose.position.y,
                        artifact_score.artifact.pose.position.z]
                new_msg = dict(
                    report_id=artifact_score.report_id,
                    artifact_type=artifact_score.artifact.type,
                    artifact_position=apos,
                    report_status=artifact_score.report_status,
                    score_change=artifact_score.score_change,
                )
                self.bus.publish('base_station', new_msg)
            else:
                self.bus.publish('radio', [addr, msg])
            return
        if data.startswith(b'points'):
            return
        if data.startswith(b'debug'):
            self.bus.publish('debug', data[6:])  # with space
            return
        frame_id = get_frame_id(data)
 #       print(frame_id)
        # TODO parse properly header "frame ID"
        #if frame_id.endswith(b'camera_depth_frame'):
        #    import pdb
        #    pdb.set_trace()
        if frame_id.endswith(b'/camera_front') or frame_id.endswith(b'/camera_down') or frame_id.endswith(b'/camera_color_optical_frame') or frame_id.endswith(b'/depth_camera_front') or frame_id.endswith(b'/rgbd_front'):
            # used to be self.topic_type == 'sensor_msgs/CompressedImage'
            self.bus.publish('image', parse_jpeg_image(packet))
        elif frame_id.endswith(b'/front_laser') or frame_id.endswith(b'/laser_front') or frame_id.endswith(b'/lidar_front'):  # self.topic_type == 'sensor_msgs/LaserScan':
            self.count += 1
            if self.count % self.downsample != 0:
                return
            self.bus.publish('scan', parse_laser(packet))
        elif frame_id.endswith(b'odom'):  #self.topic_type == 'nav_msgs/Odometry':
            __, (x, y),rot = parse_odom(packet)
            self.bus.publish('pose2d', [round(x*1000),
                                        round(y*1000),
                                        round(math.degrees(rot[0])*100)])
            self.bus.publish('t265_rot', [round(math.degrees(angle)*100)
                                     for angle in rot])

            #workaround for not existing /clock on MOBoS
            self.publish_desired_speed()

        elif frame_id.endswith(b'/base_link/imu_sensor'):  # self.topic_type == 'std_msgs/Imu':
            acc, rot, orientation = parse_imu(packet)
            self.bus.publish('rot', [round(math.degrees(angle)*100)
                                     for angle in rot])
            self.bus.publish('acc', [round(x * 1000) for x in acc])
            self.bus.publish('orientation', list(orientation))
        elif frame_id.endswith(b'/clock'):
            prev = self.timestamp_sec
            self.timestamp_sec, self.timestamp_nsec = parse_clock(packet)
            self.bus.publish('sim_clock', [self.timestamp_sec, self.timestamp_nsec//1000])  # publish us resolution (~Python.timedelta)
            if prev != self.timestamp_sec:
                self.bus.publish('sim_time_sec', self.timestamp_sec)

            ms = self.timestamp_nsec//1000000
            if self.timestamp_sec > 0 and ms % 50 == 0:  # 20Hz
                self.publish_desired_speed()
        elif frame_id.endswith(b'/gas_detected'):
            # send only status change
            if self.gas_detected != parse_bool(packet):
                self.gas_detected = parse_bool(packet)
                self.bus.publish('gas_detected', self.gas_detected)

        elif b'\0' in packet[:MAX_TOPIC_NAME_LENGTH]:
            ros_name = packet[:packet.index(b'\0')].decode('ascii')
            for row in self.topics:
                n, t = row[:2]
                new_name = n if len(row) == 2 else row[2]

                if ros_name == n:
                    result = parse_topic(t, packet[len(ros_name) + 1:])
                    if t == 'sensor_msgs/JointState':
                        name, position, velocity, effort = result

                        # publish names only on change
                        if self.joint_name != name:
                            self.bus.publish('joint_name', list(name))
                            self.joint_name = name

                        self.bus.publish('joint_position', list(position))
                        self.bus.publish('joint_velocity', list(velocity))
                        self.bus.publish('joint_effort', list(effort))
                    elif t == 'sensor_msgs/Imu':
                        acc, rot, orientation = result
                        self.bus.publish('rot', [round(math.degrees(angle)*100)
                                                 for angle in rot])  # this is deprecated and will be removed
                        self.bus.publish('acc', [round(x * 1000) for x in acc])  # potential name conflict
                        self.bus.publish(new_name, list(orientation))  # typically 'orientation'
                    elif t == 'sensor_msgs/LaserScan':
                        self.count += 1
                        if self.count % self.downsample == 0:
                            self.bus.publish(new_name, list(result))  # typically 'scan'
                    else:
                        self.bus.publish(new_name, result)

    def slot_desired_speed(self, timestamp, data):
        self.desired_speed, self.desired_angular_speed = data[0]/1000.0, math.radians(data[1]/100.0)
        self.desired_speed_3d, self.desired_angular_speed_3d = None, None

    def slot_desired_speed_3d(self, timestamp, data):
        self.desired_speed, self.desired_angular_speed = None, None
        self.desired_speed_3d, self.desired_angular_speed_3d = data

    def slot_stdout(self, timestamp, data):
        cmd = b'stdout ' + bytes(data, 'utf-8')  # redirect to ROS_INFO
        self.bus.publish('cmd', cmd)

    def slot_request_origin(self, timestamp, data):
        cmd = b'request_origin'
        self.bus.publish('cmd', cmd)

    def slot_broadcast(self, timestamp, data):
        cmd = b'broadcast ' + data  # data should be already type bytes
        self.bus.publish('cmd', cmd)

    def run(self):
        try:
            while True:
                timestamp, channel, data = self.bus.listen()
                if channel == 'raw':
                    self.slot_raw(timestamp, data)
                elif channel == 'desired_speed':
                    self.slot_desired_speed(timestamp, data)
                elif channel == 'desired_speed_3d':
                    self.slot_desired_speed_3d(timestamp, data)
                elif channel == 'stdout':
                    self.slot_stdout(timestamp, data)
                elif channel == 'request_origin':
                    self.slot_request_origin(timestamp, data)
                elif channel == 'broadcast':
                    self.slot_broadcast(timestamp, data)
                else:
                    assert False, channel  # unsupported input channel
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



"""
  Publish jpg images captured using libuvc (pupil-labs fork).
  Only cameras directly supporting jpg output are supported.
"""

from threading import Thread
import sys

import uvc


class UsbCam:
    def __init__(self, config, bus):
        self.input_thread = Thread(target=self.run_input, daemon=True)
        self.bus = bus

        bus = config['bus']
        ports = config['ports']
        # to find proper bus and ports numbers run
        # python -c "import uvc; from pprint import pprint; pprint(uvc.device_list())"

        self.sleep = config.get('sleep', 0.1)
        self.cap = None
        
        dev_list = uvc.device_list()
        for dev in dev_list:
            if dev['bus_number'] == bus and dev['bus_ports'] == ports:
                self.cap = uvc.Capture(dev["uid"])
                self.cap.frame_mode = self.cap.avaible_modes[0]
                self.cap.bandwidth_factor = 1.2
                # when incomplete pictures are being received, increase bandwidth_factor
                # when "no space left on device" error, decrease bandwidth_factor
                # for more info see https://docs.pupil-labs.com/#jpeg-size-estimation-and-custom-video-backends
                return

        import pprint
        error = "camera not found on bus %i and ports %s\n" % (bus, str(ports))
        error += "avalable cameras:\n"
        error += pprint.pformat(uvc.device_list())
        self.bus.report_error(error)
        print(error, file=sys.stderr)

    def start(self):
        self.input_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)

    def run_input(self):
        if self.cap is None:
            return
        while self.bus.is_alive():
            frame = self.cap.get_frame_robust()
            self.bus.publish('jpg', bytes(frame.jpeg_buffer))
            self.bus.sleep(self.sleep)
        self.cap.close()

    def request_stop(self):
        self.bus.shutdown()


if __name__ == "__main__":
    from pprint import pprint
    pprint(uvc.device_list())

# vim: expandtab sw=4 ts=4




import unittest

from subt.trace import Trace

class Test(unittest.TestCase):

    def test_trace(self):
        st = Trace()
        self.assertEqual(len(st.trace), 0)  # unknown initial location

        st.update_trace((0, 0, 0))
        self.assertEqual(len(st.trace), 1)  # initialized with (0, 0, 0)

        st.update_trace((0, 0, 0))
        self.assertEqual(len(st.trace), 1)  # duplicities removed

        st.update_trace((1.0, 0, 0))
        self.assertEqual(len(st.trace), 2)

        # now move in 3D up
        st.update_trace((1.0, 0, 1.0))
        self.assertEqual(len(st.trace), 3)

        # and down ...
        st.update_trace((1.0, 0, 0.0))
        self.assertEqual(len(st.trace), 4)

        st.prune()
        self.assertEqual(len(st.trace), 2)

        # now longer loop
        for i in range(10):
            st.update_trace((1, i, 0))

        for i in range(10):
            st.update_trace((1, 9 - i, 0))

        st.update_trace((2, 0, 0))
        st.update_trace((3, 0, 0))
        self.assertEqual(len(st.trace), 22)

        st.prune()
        self.assertEqual(len(st.trace), 4)

    def test_add_line_to(self):
        t = Trace(1.0)
        t.add_line_to((0, 0, 0))
        t.add_line_to((10, 0, 0))
        self.assertEqual(t.trace[-1], (10, 0, 0))
        self.assertEqual(len(t.trace), 11)

    def test_start_position(self):
        t = Trace(1.0)
        self.assertIsNone(t.start_position())
        t.update_trace((10, 20, 30))  # the start does not have to be (0, 0, 0) any more
        self.assertEqual(t.start_position(), (10, 20, 30))




#
#  Multi-stream logging with timestamps
#
# This is a simple logger for storing several independent data streams. It was
# developed during Naio competition (see https://github.com/robotika/naio) as
# an alternative to multiple log files. The concept is similar to ADTF (Automotive
# Data and Time-Triggered Framework) DAT files.
#
#   The file is using a hand made serialization of Python timedelta structures
# for timestamps.
# Each data block has header of form:
#   <timestamp>, <streamID>, <size of data>, <data>
# where "timestamp" is 32bit unsigned integer in microseconds, "streamID" is
# 16bit unsigned integer defining stream (0 is reserved for info, input and
# output have different ID), and "size" is 16bit unsigned integer defining
# the size of followed data.
#
#   The choice of microseconds was compromise between necessary number of bytes
# for timestamp and available resolution. 32bit corresponds to 4294 seconds, i.e.
# more than an hour of recording. The file contains absolute time and date in
# the overall file header allowing splitting the file and getting absolute time
# from deltas if necessary.
#
#   The stream ID is currently just integer without detailed description. There
# is planned extra info stored in "info channel/stream" ID = 0.
#
#   Finally limiting the block size to 65kB can be overcome by pre-reserved value
# 0 and 0xFFFF. The big block could be then split into several smaller once. This
# part is not defined yet.
#

import datetime
import struct
import os
import logging
import time
from threading import RLock
from ast import literal_eval
import mmap
import pathlib

g_logger = logging.getLogger(__name__)

MAGIC = b'Pyr\x00'

INFO_STREAM_ID = 0
ENV_OSGAR_LOGS = 'OSGAR_LOGS'

TIMESTAMP_OVERFLOW_STEP = (1 << 32)  # in microseconds resolution
TIMESTAMP_MASK = TIMESTAMP_OVERFLOW_STEP - 1


def format_header(start_time):
    t = start_time
    ret = []
    ret.append(MAGIC)
    ret.append(struct.pack('HBBBBBI', t.year, t.month, t.day, t.hour, t.minute, t.second, t.microsecond))
    return ret


def format_packet(stream_id, data, dt):
    ret = []
    dt_bytes = format_timedelta(dt)
    index = 0
    while index + 0xFFFF <= len(data):
        ret.append(dt_bytes + struct.pack('HH', stream_id, 0xFFFF))  # header
        ret.append(data[index:index + 0xFFFF]) # partial data
        index += 0xFFFF
    ret.append(dt_bytes + struct.pack('HH', stream_id, len(data) - index))  # header
    ret.append(data[index:])
    return ret


def format_timedelta(dt):
    assert dt.days == 0, dt  # multiple days not supported yet
    time_frac = (dt.seconds * 1000000 + dt.microseconds) & TIMESTAMP_MASK
    return time_frac.to_bytes(4, 'little')


def timedelta_parser(start_time=datetime.timedelta()):
    """ create timedelta parser

    Created parser converts `uint32` microseconds to `timedelta` object. The parser internally
    tracks overflows in `us_offset` and `prev_microseconds`. The intended use is to sequentially
    parse incoming timestamps. The argument `start_time` supports creation  of a parser where
    the last processed timestamp corresponds to the given timedelta object.
    """
    times, prev = divmod(start_time, datetime.timedelta(microseconds=TIMESTAMP_OVERFLOW_STEP))
    us_offset = times * TIMESTAMP_OVERFLOW_STEP
    prev_microseconds = prev.microseconds + prev.seconds * 10**6 + prev.days * 24 * 3600 * 10**6
    def parse_timedelta(data):
        nonlocal us_offset, prev_microseconds
        assert len(data) == 4
        microseconds = int.from_bytes(data, 'little')
        if prev_microseconds > microseconds:
            us_offset += TIMESTAMP_OVERFLOW_STEP
        prev_microseconds = microseconds
        return datetime.timedelta(microseconds=microseconds+us_offset)
    return parse_timedelta


class LogWriter:
    def __init__(self, prefix='', note='', filename=None, start_time=None):
        self.lock = RLock()
        if start_time is None:
            self.start_time = datetime.datetime.now(datetime.timezone.utc)
        else:
            self.start_time = start_time
        if filename is None:
            self.filename = prefix + self.start_time.strftime("%y%m%d_%H%M%S.log")
        else:
            self.filename = filename

        if not pathlib.Path(self.filename).is_absolute():
            if ENV_OSGAR_LOGS in os.environ:
                self.filename = os.path.join(os.environ[ENV_OSGAR_LOGS], self.filename)
            else:
                logging.warning('Environment variable %s is not set - using working directory' % ENV_OSGAR_LOGS)
            self.filename = str(pathlib.Path(self.filename).absolute())

        os.makedirs(pathlib.Path(self.filename).parent, exist_ok=True)
        self.f = open(self.filename, 'wb')

        self.f.write(b"".join(format_header(self.start_time)))
        self.f.flush()
        if len(note) > 0:
            self.write(stream_id=INFO_STREAM_ID, data=bytes(note, encoding='utf-8'))
        self.names = []

    def register(self, name):
        with self.lock:
            assert name not in self.names, (name, self.names)
            self.names.append(name)
            self.write(stream_id=INFO_STREAM_ID, data=bytes(str({'names': self.names}), encoding='ascii'))
            return len(self.names)

    def write(self, stream_id, data, dt=None):
        with self.lock:
            if dt is None:
                # by defaut generate timestamps automatically
                dt = datetime.datetime.now(datetime.timezone.utc) - self.start_time
            packet = format_packet(stream_id, data, dt)
            self.f.write(b"".join(packet))
            self.f.flush()
        return dt

    def close(self):
        self.f.close()
        self.f = None


    # context manager functions
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class LogReader:
    def __init__(self, filename, follow=False, only_stream_id=None):
        self.filename = filename
        self.follow = follow
        self.f = open(self.filename, 'rb')
        data = self._read(4)
        assert data == MAGIC, data

        data = self._read(12)
        self.start_time = datetime.datetime(*struct.unpack('HBBBBBI', data), datetime.timezone.utc)

        if only_stream_id is None:
            self.multiple_streams = set()
        else:
            try:
                self.multiple_streams = set(only_stream_id)
            except TypeError:
                self.multiple_streams = set([only_stream_id])

        self.gen = self._read_gen()

    def _read(self, size):
        buf = self.f.read(size)
        if self.follow:
            while len(buf) < size:
                time.sleep(0.1)
                buf += self.f.read(size - len(buf))
        return buf

    def _read_gen(self, only_stream_id=None):
        "packed generator - yields (time, stream, data)"
        parse_timedelta = timedelta_parser()
        while True:
            start = self.f.tell()
            header = self._read(8)
            if len(header) < 8:
                break
            dt_bytes = header[:4]
            dt =  parse_timedelta(dt_bytes)
            stream_id, size = struct.unpack('HH', header[4:])
            data = self._read(size)
            if len(data) != size:
                g_logger.error(f"Incomplete log file {self.f.name} from position {start}")
                return
            while size == 0xFFFF:
                header = self._read(8)
                if len(header) < 8:
                    break
                ref_dt_bytes = header[:4]
                ref_stream_id, size = struct.unpack('HH', header[4:])
                if dt_bytes != ref_dt_bytes or stream_id != ref_stream_id:
                    g_logger.error(f"Corrupted log file {self.f.name} from position {start}")
                    return
                part = self._read(size)
                assert len(part) == size, (len(part), size)
                data += part

            if len(self.multiple_streams) == 0 or stream_id in self.multiple_streams:
                yield dt, stream_id, data

    def close(self):
        self.f.close()
        self.f = None


    # context manager functions
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __next__(self):
        return next(self.gen)

    def __iter__(self):
        return self


class LogAsserter(LogReader):
    def __init__(self, filename, only_stream_id=None):
        LogReader.__init__(self, filename, only_stream_id=only_stream_id)
        self.assert_stream_id = None

    def write(self, stream_id, data):
        if self.assert_stream_id is not None:
            pass


def _create_index(data, pos, start_time=datetime.timedelta()):
    """index is list of (pos,timestamp) where the last item
       `pos` points just past the last valid packet
       `timestamp` is last valid timestamp
       this marker says where and when the next parsing
       should continue when growing the file
    """
    index = []
    end = len(data)
    parse_timedelta = timedelta_parser(start_time)
    dt = start_time
    while pos + 8 <= end:
        start = pos
        header = data[pos:pos+8]
        dt_bytes = header[:4]
        channel, size = struct.unpack('HH', header[4:])
        dt = parse_timedelta(dt_bytes)
        index.append((start, dt))
        pos += 8 + size
        while size == 0xFFFF and pos + 8 <= end:
            header = data[pos:pos+8]
            dt_bytes_ = header[:4]
            channel_, size = struct.unpack('HH', header[4:])
            if dt_bytes_ != dt_bytes or channel_ != channel:
                g_logger.error(f"Corrupted log file from position {start}")
                return index  # packet invalid
            pos += 8 + size
        if pos > end: # current packet is not complete
            return index
    index.append((pos, dt))  # eof marker - index of nonexisting packet
    return index


class LogIndexedReader:
    def __init__(self, filepath):
        self.filepath = filepath

    def __enter__(self):
        self.fd = os.open(self.filepath, os.O_RDONLY)
        self.data = mmap.mmap(self.fd, 0, access=mmap.ACCESS_READ)
        assert self.data[0:4] == b'Pyr\x00', self.data[0:4]
        start_time = datetime.datetime(*struct.unpack('HBBBBBI', self.data[4:4+12]))
        self.index = _create_index(self.data, 4+12)
        assert self.index[-1][0] <= len(self.data), (self.index[-1][0], len(self.data))
        return self

    def __exit__(self, *args):
        self.data.close()
        os.close(self.fd)

    def __getitem__(self, index):
        if abs(index) >= len(self):
            raise IndexError("log index {} out of range".format(index))
        assert index < len(self)
        a = self.data[self.index[index][0]:self.index[index+1][0]]
        pos = 0
        data = bytes()
        while True:
            micros, channel, size = struct.unpack('IHH', a[pos:pos+8])
            pos += 8
            data += a[pos:pos+size]
            pos += size
            if size < 0xFFFF:
                break
        dt = self.index[index][1]
        return dt, channel, data

    def grow(self):
        if (len(self.data) < self.data.size()):
            self.data.close()
            self.data = mmap.mmap(self.fd, 0, access=mmap.ACCESS_READ)
            index = _create_index(self.data, self.index[-1][0], self.index[-1][1])
            self.index[-1:] = index
        return len(self.index)-1

    def __len__(self):
        return len(self.index)-1


def lookup_stream_names(filename):
    names = []
    with LogReader(filename) as log:
        for __, channel, line in log:
            # optimization - all names are defined BEFORE other data on other channels
            if channel != 0:
                break
            if b'Errno' in line:
                continue
            d = literal_eval(line.decode('ascii'))
            if 'names' in d:
                names = d['names']
    return names


def lookup_stream_id(filename, stream_name):
    if stream_name is None:
        return None
    try:
        return int(stream_name)
    except ValueError:
        pass
    names = lookup_stream_names(filename)
    return names.index(stream_name) + 1

def calculate_stat(filename):
    names = ['sys'] + lookup_stream_names(filename)
    sizes = [0] * len(names)
    counts = [0] * len(names)
    with LogReader(filename) as log:
        for timestamp, stream_id, data in log:
            sizes[stream_id] += len(data)
            counts[stream_id] += 1
    return names, sizes, counts, timestamp


def main():
    import argparse
    import sys

    from osgar.lib.serialize import deserialize

    parser = argparse.ArgumentParser(description='Extract data from log')
    parser.add_argument('logfile', help='filename of stored file')
    parser.add_argument('--stream', help='stream ID or name', default=None, nargs='*')
    parser.add_argument('--list-names', '-l', help='list stream names', action='store_true')
    parser.add_argument('--sec', help='display timestamps in seconds', action='store_true')
    parser.add_argument('--format', help='use python format - available fields sec, timestamp, stream_id, data')
    parser.add_argument('--all', help='dump all messages', action='store_true')
    parser.add_argument('--raw', help='dump raw data', action='store_true')
    args = parser.parse_args()

    if args.list_names:
        print(lookup_stream_names(args.logfile))
        sys.exit()

    if args.stream is None and not args.all:
        names, sizes, counts, timestamp = calculate_stat(args.logfile)
        seconds = timestamp.total_seconds()
        sizes_s = [str(v) for v in sizes]
        counts_s = [str(v) for v in counts]
        name_w = max([len(v) for v in names])
        sizes_w = max([len(v) for v in sizes_s])
        counts_w = max([len(v) for v in counts_s])

        header = f' k {"name":>{name_w}} {"bytes":>{sizes_w}} | {"count":>{counts_w}} | {"freq"} Hz'
        print(header)
        print('-' * len(header))
        for k in range(len(names)):
            print(f'{k:2d} {names[k]:>{name_w}} {sizes_s[k]:>{sizes_w}} | {counts_s[k]:>{counts_w}} | {counts[k]/seconds:5.1f}Hz')
        print('\nTotal time', timestamp)
        sys.exit()

    if args.stream is None:
        only_stream = None
    else:
        only_stream = []
        for name in args.stream:
            only_stream.append(lookup_stream_id(args.logfile, name))

    with LogReader(args.logfile, only_stream_id=only_stream) as log:
        for timestamp, stream_id, data in log:
            if stream_id != 0:
                data = deserialize(data)
            if args.sec:
                print(timestamp.total_seconds(), stream_id, data)
            elif args.format:
                kw = dict(sec=timestamp.total_seconds(),
                          timestamp=timestamp,
                          walltime=log.start_time+timestamp,
                          stream_id=stream_id,
                          data=data,
                          )
                print(eval(f"f'{args.format}'", kw))
            elif args.raw:
                sys.stdout.buffer.write(data)
            else:
                print(timestamp, stream_id, data)

if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4



"""
  Logger for ZeroMQ communication
"""

from threading import Thread

import zmq

from osgar.bus import BusShutdownException


class LogZeroMQ:
    def __init__(self, config, bus):
        bus.register('raw:gz' if config.get('save_data', False) else 'raw:null', 'response', 'timeout')
        mode = config['mode']
        self.endpoint = config['endpoint']
        self.timeout = config.get('timeout', 1)  # default recv timeout 1s

        if mode == 'PULL':
            self.thread = Thread(target=self.run_input)
        elif mode == 'PUSH':
            self.thread = Thread(target=self.run_output)
        elif mode == 'REQ':
            self.thread = Thread(target=self.run_reqrep)
        else:
            assert False, mode  # unknown/unsupported mode

        self.thread.name = bus.name
        self.bus = bus

    def start(self):
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout=timeout)

    def run_input(self):
        context = zmq.Context()
        socket = context.socket(zmq.PULL)
        # https://stackoverflow.com/questions/7538988/zeromq-how-to-prevent-infinite-wait
        socket.RCVTIMEO = int(self.timeout * 1000)  # convert to milliseconds
        socket.connect(self.endpoint)

        while self.bus.is_alive():
            try:
                message = socket.recv()
                self.bus.publish('raw', message)
            except zmq.error.Again:
                pass
        socket.close()
        context.term()

    def run_output(self):
        context = zmq.Context()
        socket = context.socket(zmq.PUSH)
        # https://stackoverflow.com/questions/24619490/how-do-i-clear-the-buffer-upon-start-exit-in-zmq-socket-to-prevent-server-from
        # ZMQ_LINGER: Set linger period for socket shutdown
        # The default value of -1 specifies an infinite linger period.
        # Pending messages shall not be discarded after a call to
        # zmq_close(); attempting to terminate the socket's context with
        # zmq_term() shall block until all pending messages have been sent
        # to a peer.
        socket.setsockopt(zmq.LINGER, 100)  # milliseconds
        socket.connect(self.endpoint)
        try:
            while True:
                dt, __, data = self.bus.listen()
                while self.bus.is_alive():
                    try:
                        socket.send(data, zmq.NOBLOCK)
                        break
                    except zmq.error.Again:
                        self.bus.sleep(0.1)
        except BusShutdownException:
            pass
        socket.close()
        context.term()

    def run_reqrep(self):
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        socket.RCVTIMEO = int(self.timeout * 1000)  # convert to milliseconds
        socket.connect(self.endpoint)
        try:
            while True:
                dt, __, data = self.bus.listen()
                token, request = data
                socket.send_string(request)
                while self.bus.is_alive():
                    try:
                        resp = socket.recv().decode('ascii')
                        self.bus.publish('response', [token, resp])
                        break
                    except zmq.error.Again:
                        self.bus.publish('timeout', [token, True])
        except BusShutdownException:
            pass
        socket.close()
        context.term()

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



import logging
import unittest
import unittest.mock

from . import launch


class Test(unittest.TestCase):
    def test_launch(self):
        p = launch.Launch(config={'command': ['true']}, bus=unittest.mock.MagicMock())
        p.start()
        p.join()

    def test_request_stop(self):
        p = launch.Launch(config={'command': ['sleep', '10']}, bus=unittest.mock.MagicMock())
        p.start()
        p.request_stop()
        p.join(0.1)

    def test_join(self):
        with self.assertLogs(level=logging.WARNING) as log:
            p = launch.Launch(config={'command': ['sleep', '10']}, bus=unittest.mock.MagicMock())
            p.start()
            p.join(0.1)
        self.assertEqual(len(log.records), 1)

    def test_shell(self):
        with self.assertLogs(level=logging.WARNING) as log:
            p = launch.Launch(config={'command': 'sleep 10', 'shell': True}, bus=unittest.mock.MagicMock())
            p.start()
            p.join(0.1)
        self.assertEqual(len(log.records), 1)



import unittest

import numpy as np

from osgar.lib.depth import depth2dist


class DepthTest(unittest.TestCase):

    def Xtest_danger2dist(self):
        danger = np.zeros((360, 640), dtype=bool)
        danger[100][100] = True
        danger[200][630] = True
        dist = danger2dist(danger)
        self.assertEqual(len(dist), 640)
        self.assertEqual(dist[100], 100)
        self.assertEqual(dist[630], 200)

        values = danger[dist, np.arange(640)]
        self.assertTrue(values[100])
        self.assertTrue(values[630])
        self.assertFalse(values[300])

# vim: expandtab sw=4 ts=4



import unittest
import threading
import time

from unittest.mock import MagicMock
from datetime import timedelta

from osgar.bus import (Bus, BusShutdownException,
                       LogBusHandler, LogBusHandlerInputsOnly,
                       almost_equal)

from osgar.lib.serialize import serialize


class BusHandlerTest(unittest.TestCase):

    def test_publish(self):
        logger = MagicMock()
        bus = Bus(logger)
        handle = bus.handle('test')
        with self.assertRaises(KeyError):
            handle.publish('raw', b'some binary data')

        logger = MagicMock()
        logger.register = MagicMock(return_value=1)
        bus = Bus(logger)
        handle = bus.handle('test')
        handle.register('raw')
        handle.publish('raw', b'some binary data 2nd try')
        logger.write.assert_called_once_with(1, b'\xc4\x18some binary data 2nd try')

    def test_publish_serialization(self):
        logger = MagicMock()
        logger.register = MagicMock(return_value=1)
        bus = Bus(logger)
        handle = bus.handle('test')
        handle.register('position')
        handle.publish('position', (-123, 456))
        logger.write.assert_called_once_with(1, b'\x92\xd0\x85\xcd\x01\xc8')

    def test_listen(self):
        logger = MagicMock()
        bus = Bus(logger)
        handler = bus.handle('test')
        handler.queue.put((1, 2, b"bin data"))
        self.assertEqual(handler.listen(), (1, 2, b"bin data"))

    def test_two_modules(self):
        logger = MagicMock()
        bus = Bus(logger)
        handler2 = bus.handle('2')
        handler1 = bus.handle('1')
        handler1.register('raw')
        bus.connect('1.raw', '2.42')

        logger.write = MagicMock(return_value=123)
        handler1.publish('raw', b"Hello!")

        self.assertEqual(handler2.listen(), (123, '42', b"Hello!"))

    def test_shutdown(self):
        logger = MagicMock()
        bus = Bus(logger)
        handler = bus.handle('test')
        handler.shutdown()
        with self.assertRaises(BusShutdownException):
            handler.listen()

    def test_named_publish(self):
        logger = MagicMock()
        logger.register = MagicMock(return_value=1)
        bus = Bus(logger)
        handle = bus.handle('gps_serial')
        handle.register('raw')
        handle.publish('raw', b'bin data')
        logger.write.assert_called_once_with(1, b'\xc4\x08bin data')

    def test_alive(self):
        logger = MagicMock()
        bus = Bus(logger)
        handle = bus.handle('test')
        self.assertTrue(handle.is_alive())
        handle.shutdown()
        self.assertFalse(handle.is_alive())

    def test_log_bus_handler(self):
        log_data = [
            (timedelta(microseconds=10), 1, serialize(b'(1,2)')),
            (timedelta(microseconds=11), 1, serialize(b'(3,4,5)')),
            (timedelta(microseconds=30), 2, serialize([8, 9])),
        ]
        log = iter(log_data)
        inputs = {1:'raw'}
        outputs = {2:'can'}
        bus = LogBusHandler(log, inputs, outputs)
        bus.listen()
        with self.assertRaises(AssertionError) as e:
            bus.publish('can', b'parsed data')
        self.assertEqual(e.exception.args[0], (b'parsed data', [8, 9], timedelta(0, 0, 30)))

    def test_wrong_publish_channel(self):
        log_data = [
            (timedelta(microseconds=10), 1, serialize(b'(1,2)')),
            (timedelta(microseconds=30), 2, serialize(b'[8,9]')),
            (timedelta(microseconds=35), 3, serialize(b'[8,9]')),
        ]
        log = iter(log_data)
        inputs = {1:'raw'}
        outputs = {2:'can', 3:'can2'}
        bus = LogBusHandler(log, inputs, outputs)
        bus.listen()
        with self.assertRaises(AssertionError) as e:
            bus.publish('can2', [8, 9])
        self.assertEqual(e.exception.args[0], ('can2', 'can', timedelta(0, 0, 30)))

        with self.assertRaises(AssertionError) as e:
            bus.publish('can3', [1, 2])
        self.assertEqual(e.exception.args[0], ('can3', ('can', 'can2')))

    def test_log_bus_handler_inputs_onlye(self):
        log_data = [
            (timedelta(microseconds=10), 1, serialize([1, 2])),
            (timedelta(microseconds=11), 1, serialize([3, 4, 5])),
            (timedelta(microseconds=30), 2, serialize([8, 9])),
        ]
        log = iter(log_data)
        inputs = {1:'raw'}
        bus = LogBusHandlerInputsOnly(log, inputs)
        self.assertEqual(bus.listen(), (timedelta(microseconds=10), 'raw', [1, 2]))
        bus.publish('new_channel', b'some data')
        self.assertEqual(bus.listen(), (timedelta(microseconds=11), 'raw', [3, 4, 5]))

    def test_report_error(self):
        log = MagicMock()
        bus = Bus(log)
        handle = bus.handle('test')
        handle.report_error(error=KeyError(123))
        expected = bytes(str(dict(name='test', error=KeyError(123))), 'ascii')
        log.write.assert_called_once_with(0, expected)

    def test_publish_time(self):
        logger = MagicMock()
        logger.register = MagicMock(return_value=1)
        logger.write = MagicMock(return_value=timedelta(123))
        bus = Bus(logger)
        handle = bus.handle('test')
        handle.register('stdout')
        self.assertEqual(handle.publish('stdout', 'hello world!'), timedelta(123))

    def test_bus_sleep(self):
        logger = MagicMock()
        bus = Bus(logger)
        handle = bus.handle('test')
        handle.sleep(0.1)

        bus = LogBusHandler(logger, inputs={}, outputs={})
        bus.sleep(0.1)

        bus = LogBusHandlerInputsOnly(logger, inputs={})
        bus.sleep(0.1)

    def test_almost_equal(self):
        self.assertTrue(almost_equal(-0.27335569599868276, -0.2733556959986828))
        self.assertFalse(almost_equal(-0.27, 0.42))
        self.assertTrue(almost_equal(
            [[-0.27335569599868276, 2.625292235055242, -0.14962045778119396], [0.7263869464850868, 0.0064504746910471825, -0.007071867505754647, 0.6873936102882865]], 
            [[-0.2733556959986828,  2.625292235055242, -0.14962045778119396], [0.7263869464850868, 0.0064504746910471825, -0.007071867505754647, 0.6873936102882865]]
            ))
        self.assertFalse(almost_equal([1.23], []))
        self.assertFalse(almost_equal([-0.27], [0.42]))

    def test_almost_equal_inf(self):
        # bug detected in Moon project
        inf = float('inf')
        self.assertTrue(almost_equal(inf, inf))
        self.assertTrue(almost_equal([inf, 0, 10], [inf, 0, 10]))
        self.assertFalse(almost_equal(-inf, inf))
        self.assertFalse(almost_equal(inf, 42))

    def test_sleep(self):
        bus = Bus(MagicMock())
        handle = bus.handle('test')
        interval = 0.5
        def sleep():
            handle.sleep(interval)
        t = threading.Thread(target=sleep, daemon=True)
        start = time.monotonic()
        t.start()
        t.join(10)
        end = time.monotonic()
        # add extra epsilon time to avoid rounding error:
        # AssertionError: 255.85899999999998 not greater than or equal to 255.859
        eps = 0.000001  # 1us
        self.assertFalse(t.is_alive())
        self.assertGreaterEqual(end-interval+eps, start)

    def test_interruptible_sleep(self):
        bus = Bus(MagicMock())
        handle = bus.handle('test')
        def sleep():
            handle.sleep(10)
        t = threading.Thread(target=sleep, daemon=True)
        t.start()
        handle.shutdown()
        t.join(0.01)
        self.assertFalse(t.is_alive())

# vim: expandtab sw=4 ts=4



"""
  Convert 3D points to "lidar" scan
"""
import math

import cv2
import numpy as np

from osgar.node import Node
from osgar.bus import BusShutdownException


class PointsToScan(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("scan360")
        self.debug_arr = []

    def on_points(self, data):
        assert data.shape == (16, 1000, 3), data.shape
        self.debug_arr = data
        index = (np.arange(360) * (1000/360)).astype(int)
        xyz = data[8][index]  # mid index for 16 lidars
        X, Y, Z = xyz[:, 0], xyz[:, 1], xyz[:, 2]
        dist_i = (np.hypot(X, Y) * 1000).astype(int)
        mask = dist_i < 400
        dist_i[mask] = 0
        self.publish('scan360', dist_i.tolist())

    def update(self):
        channel = super().update()
        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        else:
            assert False, channel  # unsupported channel
        return channel

    def draw(self):
        import matplotlib.pyplot as plt

        data = self.debug_arr.reshape(16 * 1000, 3)
        for z_min, z_max, color in [(-100, -0.5, 'gray'), (1, 100, 'blue'), (-0.5, 1, 'red')]:
            arr_x = [x for x, y, z in data if z_min < z < z_max]
            arr_y = [y for x, y, z in data if z_min < z < z_max]
            plt.plot(arr_x, arr_y, 'o', linewidth=2, color=color)

        plt.axes().set_aspect('equal', 'datalim')
        plt.legend()
        plt.show()

# vim: expandtab sw=4 ts=4




import unittest
import math

from unittest.mock import MagicMock
from datetime import timedelta

import numpy as np

from osgar.bus import Bus
from osgar.lib import quaternion
from osgar.lib.unittest import TestCase

from subt.odoimuloc import Localization


class Test(TestCase):

    def test_origin(self):
        for origin in [[0, 0, 0], [1, 2, 3]]:
            with self.subTest(origin=origin):
                bus = Bus(MagicMock(write=MagicMock(return_value=timedelta())))
                loc = Localization({}, bus.handle('loc'))
                tester = bus.handle('tester')
                tester.register('origin', 'orientation', 'odom')
                bus.connect('loc.pose3d', 'tester.pose3d')
                bus.connect('tester.origin', 'loc.origin')
                bus.connect('tester.orientation', 'loc.orientation')
                bus.connect('tester.odom', 'loc.odom')
                loc.start()
                # these will be ignored until origin is received
                tester.publish('orientation', quaternion.from_axis_angle([0,0,1], math.radians(-180)))
                tester.publish('odom', [-100, -100, -100])
                tester.publish('odom', [-101, -100, -100])
                # first invalid origin
                tester.publish('origin', ['name', 'error'])
                # first valid origin accepted
                tester.publish('origin', ['name'] + origin + quaternion.identity())
                # second origin ignored
                tester.publish('origin', ['name'] + [-9, -9, -9] + quaternion.identity())
                # orientation and odom define the first pose3d output
                tester.publish('orientation', quaternion.identity())
                tester.publish('odom', origin)
                dt, channel, pose3d = tester.listen()
                self.assertEqual(channel, 'pose3d')
                self.assertPose3dEqual(pose3d, [origin, quaternion.identity()])
                loc.request_stop()
                loc.join()

    def test_odom(self):
        origin = np.asarray([10.0, 0.0, 0.0])
        inc = np.asarray([100, 0.0, 0.0])
        bus = Bus(MagicMock(write=MagicMock(return_value=timedelta())))
        loc = Localization({}, bus.handle('loc'))
        tester = bus.handle('tester')
        tester.register('origin', 'orientation', 'odom')
        bus.connect('loc.pose3d', 'tester.pose3d')
        bus.connect('tester.origin', 'loc.origin')
        bus.connect('tester.orientation', 'loc.orientation')
        bus.connect('tester.odom', 'loc.odom')
        loc.start()
        tester.publish('origin', ['name'] + list(origin) + quaternion.identity())
        tester.publish('orientation', quaternion.identity())
        tester.publish('odom', 0*inc)
        tester.publish('odom', 1*inc)
        tester.publish('odom', 2*inc)
        dt, channel, pose3d = tester.listen()
        self.assertEqual(channel, 'pose3d')
        self.assertPose3dEqual(pose3d, [list(origin), quaternion.identity()])
        dt, channel, pose3d = tester.listen()
        self.assertPose3dEqual(pose3d, [list(origin+inc/1000), quaternion.identity()])
        dt, channel, pose3d = tester.listen()
        self.assertPose3dEqual(pose3d, [list(origin+2*inc/1000), quaternion.identity()])
        loc.request_stop()
        loc.join()

    def test_orientation(self):
        origin = np.asarray([10.0, 0.0, 0.0])
        inc = np.asarray([100, 0.0, 0.0])
        inc_ori = np.asarray([0.0, 100, 0.0])
        orientation = quaternion.from_axis_angle([0, 0, 1], math.radians(90))
        bus = Bus(MagicMock(write=MagicMock(return_value=timedelta())))
        loc = Localization({}, bus.handle('loc'))
        tester = bus.handle('tester')
        tester.register('origin', 'orientation', 'odom')
        bus.connect('loc.pose3d', 'tester.pose3d')
        bus.connect('tester.origin', 'loc.origin')
        bus.connect('tester.orientation', 'loc.orientation')
        bus.connect('tester.odom', 'loc.odom')
        loc.start()
        tester.publish('origin', ['name'] + list(origin) + quaternion.identity())
        tester.publish('orientation', orientation)
        tester.publish('odom', 0*inc)
        tester.publish('odom', 1*inc)
        tester.publish('odom', 2*inc)
        dt, channel, pose3d = tester.listen()
        self.assertEqual(channel, 'pose3d')
        self.assertPose3dEqual(pose3d, [list(origin), orientation])
        dt, channel, pose3d = tester.listen()
        self.assertPose3dEqual(pose3d, [list(origin+inc_ori/1000), orientation])
        dt, channel, pose3d = tester.listen()
        self.assertPose3dEqual(pose3d, [list(origin+2*inc_ori/1000), orientation])
        loc.request_stop()
        loc.join()




"""
  Driver for articulated robot Kloubak
  (https://github.com/tf-czu/kloubak)
"""

import struct
import math
import ctypes
import enum

from osgar.node import Node
from osgar.bus import BusShutdownException

#WHEEL_DISTANCE = 0.475  # m
WHEEL_DISTANCE = 0.496  # m K2, can be modified by config
CENTER_AXLE_DISTANCE = 0.348  # distance from potentiometer, can be modified by config
VESC_REPORT_FREQ = 20  # was 100  # Hz
SPEED_ENC_SCALE = (33/25)*0.25 * math.pi / (4 * 3 * 60 * VESC_REPORT_FREQ)  # scale 4x found experimentally
ENC_SCALE = (33/25)*8.0/950  # TODO proper calibration (scale for large 33" wheels, old were 25")
TURNING_ANGULAR_SPEED = math.pi/8

AD_CENTER = 531 # K2, can be modified by config
AD_CALIBRATION_DEG = 45  # K2, can be modified by config
AD_RANGE = -178  # K2, can be modified by config
MAX_JOINT_ANGLE_DEG = 68
MAX_JOINT_ANGLE = math.radians(MAX_JOINT_ANGLE_DEG) # K2, can be modified by config
AD_CENTER2 = None
AD_RANGE2 = None

DOWNDROP_TOO_LONG_RAW = 1000  # in millimeters, trigger for downdrop/hole
DOWNDROP_TOO_SHORT_RAW = 200  # in millimeters, trigger for low obstacle, probably not visible by 2D lidar
DOWNDROP_STABLE_COUNT = 3  # if within the limits for given number of cycles report as bumper True/False

CAN_ID_BUTTONS = 0x1
CAN_ID_VESC_FRONT_R = 0x91
CAN_ID_VESC_FRONT_L = 0x92
CAN_ID_VESC_REAR_R = 0x93
CAN_ID_VESC_REAR_L = 0x94
CAN_ID_VESC_REAR_K3_R = 0x95
CAN_ID_VESC_REAR_K3_L = 0x96
CAN_ID_SYNC = CAN_ID_VESC_FRONT_L
CAN_ID_CURRENT = 0x70
CAN_ID_DOWNDROPS_FRONT = 0x71
CAN_ID_DOWNDROPS_REAR = 0x72
CAN_ID_JOINT_ANGLE = 0x80
CAN_ID_REGULATED_VOLTAGE = 0x81
CAN_ID_VOLTAGE = 0x82
CAN_ID_ENCODERS = 0x83

INDEX_FRONT_LEFT = 1
INDEX_FRONT_RIGHT = 0
INDEX_REAR_LEFT = 3
INDEX_REAR_RIGHT = 2
INDEX_REAR_K3_LEFT = 5
INDEX_REAR_K3_RIGHT = 4

MIN_SPEED = 0.3

#Transferring coefficient for the vesc tachometers to meters: distance = vesc_value * 0.845/100


class DriveMode(enum.Enum):
    FRONT = 1
    REAR = 2
    ALL = 3  # 4WD


def CAN_triplet(msg_id, data):
    return [msg_id, bytes(data), 0]  # flags=0, i.e basic addressing


def draw(arr, joint_arr):
    import matplotlib.pyplot as plt
    t = [a[0] for a in arr]
    values = [a[1:] for a in arr]

    f, ax = plt.subplots(2, sharex=True)

    line = ax[0].plot(t, values, '-', linewidth=2)
    ax[0].legend(line, ['request left', 'request right',
                      'enc front left', 'enc front right',
                      'enc rear left', 'enc rear right'])

    line = ax[1].plot(t, joint_arr, '-', linewidth=2)

    plt.xlabel('time (s)')
    plt.show()


def draw_enc_stat(arr):
    import matplotlib.pyplot as plt
    t = [a[0] for a in arr]
    values = [a[1:] for a in arr]

    line = plt.plot(t, values, '-', linewidth=2)

    plt.xlabel('time (s)')
    plt.show()


def draw_4wd(speed_arr, joint_arr):
    import numpy as np
    from matplotlib import pyplot as plt

    fig = plt.figure(figsize=(8,8))
    if len(speed_arr) > 0:
        speed_arr = np.array(speed_arr)
        print(speed_arr.shape)
        ax1 = fig.add_subplot(211)
        ax1.plot(speed_arr[:, 0], speed_arr[:, 1], "+r", label="front left")
        ax1.plot(speed_arr[:, 0], speed_arr[:, 2], "+g", label="front right")
        ax1.plot(speed_arr[:, 0], speed_arr[:, 3], "+b", label="rear left")
        ax1.plot(speed_arr[:, 0], speed_arr[:, 4], "+y", label="rear right")
        ax1.set_xlabel("Time (s)")
        ax1.set_ylabel("Speed (ms$^{-1}$)")
        ax1.legend()
    else:
        print("No speed data")

    if len(joint_arr) > 0:
        joint_arr = np.array(joint_arr)
        print(joint_arr.shape)
        ax2 = fig.add_subplot(212)
        ax2.plot(joint_arr[:, 0], joint_arr[:, 1], "+r", label="actual angle")
        ax2.plot(joint_arr[:, 0], joint_arr[:, 2], "+g", label="desired angle")
        ax2.set_xlabel("Time (s)")
        ax2.set_ylabel("Angle (deg)")
        ax2.legend()
    else:
        print("No joint data")

    plt.show()


def compute_desired_erpm(desired_speed, desired_angular_speed):
    scale = 1 / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)
    left = scale * (desired_speed - desired_angular_speed * WHEEL_DISTANCE / 2)
    right = scale * (desired_speed + desired_angular_speed * WHEEL_DISTANCE / 2)
    return int(round(left)), int(round(right))


def compute_desired_angle(desired_speed, desired_angular_speed):
    # The angle is computed from triangle, where one
    # side has length = radius (computed from speed and angular
    # speed) and the far side is CENTER_AXLE_DISTANCE.
    # The Kloubak part is perpendicular to the turning center.
    if abs(desired_angular_speed) < 0.000001:
        return 0.0
    radius = desired_speed/desired_angular_speed
    if abs(radius) < 0.000001:
        # this case is undefined, we can only for large angle
        return math.copysign(MAX_JOINT_ANGLE, desired_angular_speed)

    desired_angle = 2 * math.atan(CENTER_AXLE_DISTANCE / radius)
    if abs(desired_angle) > MAX_JOINT_ANGLE:
        return math.copysign(MAX_JOINT_ANGLE, desired_angle)
    return desired_angle


def calculate_wheels_speeds( desired_speed, desired_angular_speed, actual_angle ):
    desired_angle = compute_desired_angle( desired_speed, desired_angular_speed)
    if abs(actual_angle) < 0.05:
        v_fl = v_rl = v_fr = v_rr = desired_speed
    else:
        actual_radius = CENTER_AXLE_DISTANCE / math.tan(actual_angle/2)
        v_fl = v_rl = desired_speed / actual_radius * (actual_radius - WHEEL_DISTANCE / 2)
        v_fr = v_rr = desired_speed / actual_radius * (actual_radius + WHEEL_DISTANCE / 2)
    if abs( desired_angle - actual_angle ) < 0.05: # in radians cca 2.9 deg
        return v_fl, v_fr, v_rl, v_rr
    turning_angular_speed = min(math.pi / 3, abs(desired_angle - actual_angle) * 1 + math.pi / 16)
#    speed_correction = CENTER_AXLE_DISTANCE * TURNING_ANGULAR_SPEED * math.tan(actual_angle / 2)
    speed_correction = CENTER_AXLE_DISTANCE * turning_angular_speed * math.tan(actual_angle/2)  # positive for left, negative for right
    turning_wheel_speed = turning_angular_speed * WHEEL_DISTANCE / 2
    if desired_angle - actual_angle > 0:  # turn left
        v_fl = v_fl - turning_wheel_speed - speed_correction
        v_fr = v_fr + turning_wheel_speed - speed_correction
        v_rl = v_rl + turning_wheel_speed + speed_correction
        v_rr = v_rr - turning_wheel_speed + speed_correction
    else:  # turn right
        v_fl = v_fl + turning_wheel_speed + speed_correction
        v_fr = v_fr - turning_wheel_speed + speed_correction
        v_rl = v_rl - turning_wheel_speed - speed_correction
        v_rr = v_rr + turning_wheel_speed - speed_correction
#    print(desired_speed, desired_angular_speed, actual_angle, v_fl, v_fr, v_rl, v_rr)
    return v_fl, v_fr, v_rl, v_rr


def joint_rad(analog, second_ang = False):
    if analog is None:
        return None
    if second_ang:
        return math.radians(AD_CALIBRATION_DEG * (AD_CENTER2 - analog) / AD_RANGE2)
    return math.radians(AD_CALIBRATION_DEG * ( AD_CENTER - analog )/AD_RANGE)


def joint_deg(analog):
    ret = joint_rad(analog)
    if ret is None:
        return None
    return math.degrees(ret)


def compute_rear(speed, angular_speed, joint_angle):
    ca, sa = math.cos(joint_angle), math.sin(joint_angle)
    x = ca * speed + sa * angular_speed * CENTER_AXLE_DISTANCE
    y = -sa * speed + ca * angular_speed * CENTER_AXLE_DISTANCE
    return x, -y


def sint16_diff(a, b):
    return ctypes.c_int16(a - b).value


def setup_global_const(config):
    global WHEEL_DISTANCE
    global CENTER_AXLE_DISTANCE
    global AD_CENTER
    global AD_CENTER2
    global AD_CALIBRATION_DEG
    global AD_RANGE
    global AD_RANGE2
    global MAX_JOINT_ANGLE
#    global TURNING_WHEEL_SPEED
    WHEEL_DISTANCE = config.get("wheel_distance", WHEEL_DISTANCE )
    CENTER_AXLE_DISTANCE = config.get("center_axle_distance", CENTER_AXLE_DISTANCE )
    AD_CENTER = config.get("ad_center", AD_CENTER )
    AD_CENTER2 = config.get("ad_center2")
    AD_CALIBRATION_DEG = config.get("ad_calibration_deg", AD_CALIBRATION_DEG )
    AD_RANGE = config.get("ad_range", AD_RANGE )
    AD_RANGE2 = config.get("ad_range2")
    MAX_JOINT_ANGLE = math.radians( config.get("max_joint_angle_deg", MAX_JOINT_ANGLE_DEG) )
#    TURNING_WHEEL_SPEED = TURNING_ANGULAR_SPEED * WHEEL_DISTANCE / 2


def get_downdrop_bumpers(downdrops):
    """Create Boolean array based on raw downdrops readings"""
    return [not(DOWNDROP_TOO_SHORT_RAW < dist_mm < DOWNDROP_TOO_LONG_RAW) for dist_mm in downdrops]


class RobotKloubak(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('pose2d', 'emergency_stop', 'encoders', 'can', 'joint_angle',
                     'bumpers_front', 'bumpers_rear',
                     'downdrops_front', 'downdrops_rear')
        setup_global_const(config)

        # commands
        self.desired_speed = 0.0  # m/s
        self.desired_angular_speed = 0.0
        self.v_fl = self.v_fr = self.v_rl = self.v_rr = 0  # values in m/s
        self.drive_mode = DriveMode[config.get("drive_mode", "FRONT")]
        self.num_axis = config.get("num_axis", 2)

        # status
        self.emergency_stop = None  # unknown state
        self.pose = (0.0, 0.0, 0.0)  # x, y in meters, heading in radians (not corrected to 2PI)
        self.buttons = None
        self.last_encoders_front_left = None
        self.last_encoders_front_right = None
        self.last_encoders_rear_left = None
        self.last_encoders_rear_right = None
        self.last_encoders_rear_k3_l = None
        self.last_encoders_rear_k3_r = None
        self.last_encoders_time = None
        self.last_pose_encoders = [0, 0,] * self.num_axis  # accumulated tacho readings
        self.encoders = [0, 0,] * self.num_axis  # handling 16bit integer overflow
        self.last_encoders_16bit = None  # raw readings
        self.last_joint_angle = None  # joint angle in radians
        self.last_joint_angle2 = None  # # joint2 angle in radians
        self.prev_joints_angle = None  # angle between front and rear axle, used for heading
        self.voltage = None
        self.downdrops_front = None  # array of downdrop sensors
        self.downdrops_rear = None
        self.can_errors = 0  # count errors instead of assert

        # new per axis encoders
        self.epoch = None  # unknown
        self.partial_encoders = [None, None,] * self.num_axis

        self.verbose = False  # should be in Node
        self.enc_debug_arr = []
        self.joint_debug_arr = []
        self.speed_debug_arr = []
        self.count = [0, 0,] * self.num_axis
        self.count_arr = []
        self.debug_odo = []
        self.debug_downdrops_front = []
        self.debug_downdrops_rear = []
        self.bumpers_front = None  # unknown, report difference
        self.bumpers_front_counter = 0  # how many cycles is the True/False value stable
        self.bumpers_rear = None
        self.bumpers_rear_counter = 0
        print('Kloubak mode:', self.drive_mode, 'num_axis:', self.num_axis)

    def send_pose(self):
        x, y, heading = self.pose
        self.publish('pose2d', [round(x*1000), round(y*1000),
                                round(math.degrees(heading)*100)])

    def update_buttons(self, data):
        # assert len(data) == 1, len(data)
        if len(data) != 1:
            self.can_errors += 1
            return
        val = data[0]
        if self.buttons is None or val != self.buttons:
            self.buttons = val
            stop_status = self.buttons & 0x01 == 0x01
            if self.emergency_stop != stop_status:
                self.emergency_stop = stop_status
                self.bus.publish('emergency_stop', self.emergency_stop)
                print('Emergency STOP:', self.emergency_stop)

    def partial_axis_encoders_update(self, arr):
        # return True when completed
        # simple report encoders on epoch change -> delayed 1 cycle
        epoch, axis, left, right = arr
        ret = self.partial_encoders[:]
#        assert axis in [1, 2, 3], axis
        if axis == 0:
            print("Error: nonvalid axis: ", axis)
            return None
        self.partial_encoders[2 * (axis - 1)] = left
        self.partial_encoders[2 * (axis - 1) + 1] = right
        if self.epoch is None or self.epoch == epoch:
            self.epoch = epoch
            return None
        self.epoch = epoch
        if None in ret:
            return None  # do not publish incomplete readings
        return ret

    def update_encoders(self, msg_id, data):
        if msg_id == 0x83:
            if len(data) == 8:
                # old version used on Kloubak K2, two encoders in single message
                encoders = struct.unpack('>HHHH', data)
            elif len(data) == 6:
                # new version for Kloubak K3, one message per axis
                arr = struct.unpack('>BBHH', data)
                encoders = self.partial_axis_encoders_update(arr)
                if encoders is None:
                    return
            else:
                self.can_errors += 1
                return

            if self.last_encoders_time is not None:
                diff = [sint16_diff(e, prev) for prev, e in zip(self.last_encoders_16bit, encoders)]
            else:
                diff = [0,] * len(encoders)  # support both K2 and K3
            self.encoders = [e + d for e, d in zip(self.encoders, diff)]
            self.last_encoders_16bit = encoders
            self.last_encoders_time = self.time
            if self.verbose:
                print(self.time, diff, self.encoders)
            return encoders
        assert msg_id in [0x91, 0x92, 0x93, 0x94, 0x95, 0x96], hex(msg_id)
        # assert len(data) == 8, data
        if len(data) != 8:
            self.can_errors += 1
            return
        rpm3, current, duty_cycle = struct.unpack('>ihh', data)
        if msg_id == CAN_ID_VESC_FRONT_L:
            self.last_encoders_front_left = rpm3
#            print('left', rpm3, current, duty_cycle)
        elif msg_id == CAN_ID_VESC_FRONT_R:
            self.last_encoders_front_right = rpm3
#            print('right', rpm3, current, duty_cycle)
        if msg_id == CAN_ID_VESC_REAR_L:
            self.last_encoders_rear_left = rpm3
        elif msg_id == CAN_ID_VESC_REAR_R:
            self.last_encoders_rear_right = rpm3
        if msg_id == CAN_ID_VESC_REAR_K3_L:
            self.last_encoders_rear_k3_l = rpm3
        elif msg_id == CAN_ID_VESC_REAR_K3_R:
            self.last_encoders_rear_k3_r = rpm3
        self.count[msg_id - 0x91] += 1
#        print(self.count)
        if self.verbose:
            self.count_arr.append([self.time.total_seconds()] + self.count)

    def compute_pose(self, left, right, joint_angle_diff = 0):
        """Update internal pose with 'dt' step"""
        if left is None or right is None:
            return False, None, None
        x, y, heading = self.pose

        metricL = ENC_SCALE * left  # dt is already part of ENC_SCALE
        metricR = ENC_SCALE * right

        dist = (metricL + metricR)/2.0
        angle = (metricR - metricL)/WHEEL_DISTANCE + joint_angle_diff

        # advance robot by given distance and angle
        if abs(angle) < 0.0000001:  # EPS
            # Straight movement - a special case
            x += dist * math.cos(heading)
            y += dist * math.sin(heading)
            #Not needed: heading += angle
        else:
            # Arc
            r = dist / angle
            x += -r * math.sin(heading) + r * math.sin(heading + angle)
            y += +r * math.cos(heading) - r * math.cos(heading + angle)
            heading += angle # not normalized
        return True, (x, y, heading), (dist, angle)

    def update_pose(self):
        diff = [e - prev for e, prev in zip(self.encoders, self.last_pose_encoders)]
        self.last_pose_encoders = self.encoders
        if self.num_axis == 2:
            if self.last_joint_angle is None:
                return False
            if self.prev_joints_angle is None:
                self.prev_joints_angle = self.last_joint_angle
                return False
            joint_angle_diff = self.last_joint_angle - self.prev_joints_angle
            self.prev_joints_angle = self.last_joint_angle
            if self.desired_speed >= 0:
                ret, pose, motion = self.compute_pose(diff[INDEX_REAR_LEFT], diff[INDEX_REAR_RIGHT], joint_angle_diff)
            else:
                ret, pose, motion = self.compute_pose(diff[INDEX_FRONT_LEFT], diff[INDEX_FRONT_RIGHT])
        elif self.num_axis == 3:
            if self.last_joint_angle is None or self.last_joint_angle2 is None:
                return False
            if self.prev_joints_angle is None:
                self.prev_joints_angle = self.last_joint_angle + self.last_joint_angle2
                return False
            joint_angle_diff = self.last_joint_angle + self.last_joint_angle2 - self.prev_joints_angle
            self.prev_joints_angle = self.last_joint_angle + self.last_joint_angle2
            if self.desired_speed >= 0:
                ret, pose, motion = self.compute_pose(diff[INDEX_REAR_K3_LEFT], diff[INDEX_REAR_K3_RIGHT], joint_angle_diff)
            else:
                ret, pose, motion = self.compute_pose(diff[INDEX_FRONT_LEFT], diff[INDEX_FRONT_RIGHT])
        else:
            assert False, self.num_axis  # 2 or 3 are supported only

        if ret:
            self.pose = pose

        if self.verbose and ret and self.last_joint_angle is not None:
#            self.debug_odo.append((self.time.total_seconds(), motion[0], motion_rear[0]))
#            self.debug_odo.append((self.time.total_seconds(), motion[1], motion_rear[1]))
#            self.debug_odo.append((self.time.total_seconds(), motion[0], motion[1],
#                                   motion_rear[0], motion_rear[1],
#                                   math.degrees(self.last_joint_angle)))
            estimate = compute_rear(motion[0], motion[1], self.last_joint_angle)
#            self.debug_odo.append((self.time.total_seconds(), motion[0], motion_rear[0], estimate[0]))
#            self.debug_odo.append((self.time.total_seconds(), motion[1], motion_rear[1], estimate[1]))
#            self.debug_odo.append((self.time.total_seconds(), motion_rear[1], estimate[1]))
        return ret

    def process_packet(self, packet, verbose=False):
        msg_id, payload, flags = packet
        if msg_id == CAN_ID_BUTTONS:
            self.update_buttons(payload)
        elif msg_id == CAN_ID_CURRENT:
            # expected 24bit integer miliAmps
            if len(payload) == 3:
                current = struct.unpack('>i', b'\x00' + payload)[0]
#                print(current)
            else:
                self.can_errors += 1
        elif msg_id == CAN_ID_JOINT_ANGLE:
            if len(payload) == 4:
                if self.num_axis == 3:
                    analog_joint_angle, analog_joint_angle2 = struct.unpack('>hh', payload)
    #                print(self.last_joint_angle,payload)
#                    print(self.last_joint_angle, self.last_joint_angle2)
                    self.last_joint_angle, self.last_joint_angle2 = joint_rad(analog_joint_angle), joint_rad(analog_joint_angle2, second_ang=True)
                    self.publish('joint_angle', [round(math.degrees(self.last_joint_angle)*100), round(math.degrees(self.last_joint_angle2)*100)])
                else:
                    analog_joint_angle = struct.unpack('>i', payload)[0]
#                    print(self.last_joint_angle,payload)
#                    print(self.last_joint_angle)
                    self.last_joint_angle = joint_rad(analog_joint_angle)
                    self.publish('joint_angle', [round(math.degrees(self.last_joint_angle)*100)])
            else:
                self.can_errors += 1
        elif msg_id == CAN_ID_VOLTAGE:
            if len(payload) == 4:
                self.voltage = struct.unpack_from('>i', payload)[0]
#                print(self.voltage)
            else:
                self.can_errors += 1
        elif msg_id == CAN_ID_DOWNDROPS_FRONT:
            # assert len(packet) == 2 + 4, len(packet)
            if len(payload) == 4:
                self.downdrops_front = struct.unpack_from('>HH', payload)
                self.publish('downdrops_front', list(self.downdrops_front))
                prev = self.bumpers_front
                self.bumpers_front = get_downdrop_bumpers(self.downdrops_front)
                if prev != self.bumpers_front:
                    self.bumpers_front_counter = 0  # reset on change
                self.bumpers_front_counter += 1
                if self.bumpers_front_counter == DOWNDROP_STABLE_COUNT:  # report only "slightly stable" values
                    self.publish('bumpers_front', self.bumpers_front)
#                print(self.time, self.downdrops_front)
                if self.verbose:
                    self.debug_downdrops_front.append(
                            (self.time.total_seconds(), self.downdrops_front[0], self.downdrops_front[1]))
            else:
                self.can_errors += 1
        elif msg_id == CAN_ID_DOWNDROPS_REAR:
            # assert len(packet) == 2 + 4, len(packet)
            if len(payload) == 4:
                self.downdrops_rear = struct.unpack_from('>HH', payload)
                self.publish('downdrops_rear', list(self.downdrops_rear))
                prev = self.bumpers_rear
                self.bumpers_rear = get_downdrop_bumpers(self.downdrops_rear)
                if prev != self.bumpers_rear:
                    self.bumpers_rear_counter = 0  # reset on change
                self.bumpers_rear_counter += 1
                if self.bumpers_rear_counter == DOWNDROP_STABLE_COUNT:  # report only "slightly stable" values
                    self.publish('bumpers_rear', self.bumpers_rear)
#                print(self.time, self.downdrops_rear)
                if self.verbose:
                    self.debug_downdrops_rear.append(
                            (self.time.total_seconds(), self.downdrops_rear[0], self.downdrops_rear[1]))
            else:
                self.can_errors += 1
        elif msg_id in [CAN_ID_ENCODERS, CAN_ID_VESC_FRONT_L, CAN_ID_VESC_FRONT_R, CAN_ID_VESC_REAR_L, CAN_ID_VESC_REAR_R, CAN_ID_VESC_REAR_K3_R, CAN_ID_VESC_REAR_K3_L]:
            if self.update_encoders(msg_id, payload) is not None:
                diff = [e - prev for e, prev in zip(self.encoders, self.last_pose_encoders)]
                # note, that this craziness is necessary only because the motor indexes
                # are assigned first right and then left
                if len(diff) == 4:
                    self.publish('encoders',
                            [diff[INDEX_FRONT_LEFT], diff[INDEX_FRONT_RIGHT],
                             diff[INDEX_REAR_LEFT], diff[INDEX_REAR_RIGHT]])
                else:
                    assert len(diff) == 6, len(diff)  # Kloubak K3
                    self.publish('encoders',
                            [diff[INDEX_FRONT_LEFT], diff[INDEX_FRONT_RIGHT],
                             diff[INDEX_REAR_LEFT], diff[INDEX_REAR_RIGHT],
                             diff[INDEX_REAR_K3_LEFT], diff[INDEX_REAR_K3_RIGHT]])
                if self.update_pose():
                    self.send_pose()
                return True
        return False


    def update_drive_three_axles(self, timestamp, data):
        if self.last_joint_angle and self.desired_speed > 0:
            angle = self.last_joint_angle
            assert abs(angle) < math.pi/2, angle * 180 / math.pi
            self.v_fl, self.v_fr, self.v_rl, self.v_rr = calculate_wheels_speeds(self.desired_speed,
                                                                                 self.desired_angular_speed, angle)
            if self.verbose:
                self.joint_debug_arr.append([timestamp.total_seconds(), angle*180/math.pi, compute_desired_angle( self.desired_speed, self. desired_angular_speed )*180/math.pi] )

        elif self.last_joint_angle2 and self.desired_speed < 0:
            angle = self.last_joint_angle2
            assert abs(angle) < math.pi / 2, angle * 180 / math.pi
            self.v_fl, self.v_fr, self.v_rl, self.v_rr = calculate_wheels_speeds(self.desired_speed,
                                                                                 self.desired_angular_speed, angle)
            if self.verbose:
                self.joint_debug_arr.append([timestamp.total_seconds(), angle*180/math.pi, compute_desired_angle( self.desired_speed, self. desired_angular_speed )*180/math.pi] )

        if self.process_packet(data):
            stop = [0, 0, 0, 0]

            if self.verbose:
                self.speed_debug_arr.append( [ timestamp.total_seconds(), self.v_fl, self.v_fr, self.v_rl, self.v_rr ] )

            if self.desired_speed > 0:
                if self.last_encoders_front_right is not None:
                    v_fr_2 = int(round(self.v_fr / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x31, list(struct.pack('>i', v_fr_2))))  # front right
                if self.last_encoders_front_left is not None:
                    v_fl_2 = int(round(self.v_fl / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x32, list(struct.pack('>i', v_fl_2))))  # front left
                if self.last_encoders_rear_right is not None:
                    v_rr_2 = int(round(self.v_rr / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x33, list(struct.pack('>i', v_rr_2))))  # right mid
                if self.last_encoders_rear_left is not None:
                    v_rl_2 = int(round(self.v_rl / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x34, list(struct.pack('>i', v_rl_2))))  # left mid
                self.publish('can', CAN_triplet(0x15, stop))  # right rear
                self.publish('can', CAN_triplet(0x16, stop))  # left rear

            elif self.desired_speed < 0:
                self.publish('can', CAN_triplet(0x11, stop))  # right front
                self.publish('can', CAN_triplet(0x12, stop))  # left front

                if self.last_encoders_rear_right is not None:
                    v_fr_2 = int(round(self.v_fr / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x33, list(struct.pack('>i', v_fr_2))))  # mid right
                if self.last_encoders_rear_left is not None:
                    v_fl_2 = int(round(self.v_fl / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x34, list(struct.pack('>i', v_fl_2))))  # mid left
                if self.last_encoders_rear_k3_r is not None:
                    v_rr_2 = int(round(self.v_rr / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x35, list(struct.pack('>i', v_rr_2))))  # right rear
                if self.last_encoders_rear_k3_l is not None:
                    v_rl_2 = int(round(self.v_rl / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x36, list(struct.pack('>i', v_rl_2))))  # left rear

            else:
                self.publish('can', CAN_triplet(0x11, stop))  # right front
                self.publish('can', CAN_triplet(0x12, stop))  # left front
                self.publish('can', CAN_triplet(0x13, stop))  # right mid
                self.publish('can', CAN_triplet(0x14, stop))  # left mid
                self.publish('can', CAN_triplet(0x15, stop))  # right rear
                self.publish('can', CAN_triplet(0x16, stop))  # left rear


    def update_drive_two_axles(self, timestamp, data):
        """Drive all motors"""
        # update commands for all 4 motors only with new angle reading
        # (Kloubak does not have any SYNC signal for all motor drivers)
        # TODO integration front driver?
        if self.last_joint_angle:
            angle = self.last_joint_angle
            assert abs(angle) < math.pi / 2, angle * 180 / math.pi
            self.v_fl, self.v_fr, self.v_rl, self.v_rr = calculate_wheels_speeds(self.desired_speed,
                                                                                 self.desired_angular_speed, angle)
            if self.verbose:
                self.joint_debug_arr.append([timestamp.total_seconds(), angle*180/math.pi, compute_desired_angle( self.desired_speed, self. desired_angular_speed )*180/math.pi] )

        if self.process_packet(data):
            stop = [0, 0, 0, 0]
#
            if self.verbose:
                self.speed_debug_arr.append( [ timestamp.total_seconds(), self.v_fl, self.v_fr, self.v_rl, self.v_rr ] )

            if self.desired_speed != 0: # and self.drive_mode == DriveMode.ALL: ?
                if self.last_encoders_front_right is not None:
                    v_fr_2 = int(round(self.v_fr / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x31, list(struct.pack('>i', v_fr_2))))  # front right
                if self.last_encoders_front_left is not None:
                    v_fl_2 = int(round(self.v_fl / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x32, list(struct.pack('>i', v_fl_2))))  # front left
                if self.last_encoders_rear_right is not None:
                    v_rr_2 = int(round(self.v_rr / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x33, list(struct.pack('>i', v_rr_2))))  # rear right
                if self.last_encoders_rear_left is not None:
                    v_rl_2 = int(round(self.v_rl / (VESC_REPORT_FREQ * SPEED_ENC_SCALE)))
                    self.publish('can', CAN_triplet(0x34, list(struct.pack('>i', v_rl_2))))  # rear left

#            elif self.drive_mode == DriveMode.FRONT ?

            else:
                self.publish('can', CAN_triplet(0x11, stop))  # right front
                self.publish('can', CAN_triplet(0x12, stop))  # left front
                self.publish('can', CAN_triplet(0x13, stop))  # right rear
                self.publish('can', CAN_triplet(0x14, stop))  # left rear


    def update_drive_front_wheels(self, timestamp, data):
        send_speed = self.desired_speed
        if self.last_joint_angle is not None and abs(self.desired_speed) > MIN_SPEED:
            desired_angle = compute_desired_angle(self.desired_speed, self.desired_angular_speed)
            angle = self.last_joint_angle
            angle_diff = abs( desired_angle - angle )
            if angle_diff > 0.2 and self.desired_speed > 0:
                send_speed = max(MIN_SPEED, self.desired_speed * (1.2 - angle_diff))  # 1 rad is 57 deg
            elif angle_diff > 0.2 and self.desired_speed < 0:
                send_speed = min(- MIN_SPEED, self.desired_speed * (1.2 - angle_diff))  # 1 rad is 57 deg

        limit_l, limit_r = compute_desired_erpm(send_speed, self.desired_angular_speed)
#        limit_l, limit_r = compute_desired_erpm(self.desired_speed, self.desired_angular_speed)
        if self.process_packet(data):
            stop = [0, 0, 0, 0]
            if self.verbose:
                cmd_l, cmd_r = limit_l, limit_r
                if self.desired_speed < 0:
                    cmd_l, cmd_r = -cmd_l, -cmd_r
                elif self.desired_speed == 0:
                    cmd_l, cmd_r = 0, 0
                self.enc_debug_arr.append((timestamp.total_seconds(), cmd_l, cmd_r,
                    self.last_encoders_front_left, self.last_encoders_front_right,
                    self.last_encoders_rear_left, self.last_encoders_rear_right))
                self.joint_debug_arr.append(math.degrees(self.last_joint_angle))

            rear_drive = False  # True  # experimental
            if rear_drive and self.desired_speed > 0:
                # control the joint angle and the desired speed
                self.publish('can', CAN_triplet(0x11, stop))  # right front
                self.publish('can', CAN_triplet(0x12, stop))  # left front

                if self.last_joint_angle is not None:
                    desired_angle = compute_desired_angle(self.desired_speed, self.desired_angular_speed)
                    angle = self.last_joint_angle
                    diff = abs(desired_angle - angle)
                    speed = self.desired_speed * (1 - diff/math.radians(80))

                    angular_speed = desired_angle - angle  # i.e. in 1s it should be the same
                    limit_l, limit_r = compute_desired_erpm(speed, -angular_speed)  # mirror flip (rear)

                if self.last_encoders_rear_right is not None:
                    self.publish('can', CAN_triplet(0x33, list(struct.pack('>i', limit_r))))
                if self.last_encoders_rear_left is not None:
                    self.publish('can', CAN_triplet(0x34, list(struct.pack('>i', limit_l))))

            elif self.desired_speed > 0:
                if self.last_encoders_front_right is not None:
                    self.publish('can', CAN_triplet(0x31, list(struct.pack('>i', limit_r))))
                if self.last_encoders_front_left is not None:
                    self.publish('can', CAN_triplet(0x32, list(struct.pack('>i', limit_l))))
                self.publish('can', CAN_triplet(0x13, stop))  # rear right
                self.publish('can', CAN_triplet(0x14, stop))  # rear left

            elif self.desired_speed < 0:
                if self.last_encoders_rear_right is not None:
                    self.publish('can', CAN_triplet(0x33, list(struct.pack('>i', limit_r))))
                if self.last_encoders_rear_left is not None:
                    self.publish('can', CAN_triplet(0x34, list(struct.pack('>i', limit_l))))
                self.publish('can', CAN_triplet(0x11, stop))  # right front
                self.publish('can', CAN_triplet(0x12, stop))  # left front

            else:
                self.publish('can', CAN_triplet(0x11, stop))  # right front
                self.publish('can', CAN_triplet(0x12, stop))  # left front
                self.publish('can', CAN_triplet(0x13, stop))  # right rear
                self.publish('can', CAN_triplet(0x14, stop))  # left rear


    def slot_can(self, timestamp, data):
        self.time = timestamp
        if self.num_axis == 3:
            self.update_drive_three_axles(timestamp, data)

        elif self.num_axis == 2:
            if self.drive_mode == DriveMode.ALL:
                self.update_drive_two_axles(timestamp, data)
            elif self.drive_mode == DriveMode.FRONT:
                self.update_drive_front_wheels(timestamp, data)
            else:
                assert False, self.drive_mode  # is not supported yet

        else:
            assert False, self.num_axis  # 2 or 3 are supported only


    def slot_desired_speed(self, timestamp, data):
        self.desired_speed, self.desired_angular_speed = data[0]/1000.0, math.radians(data[1]/100.0)

    def run(self):
        try:
            while True:
                try:
                    self.time, channel, data = self.listen()
                except StopIteration:
                    if self.verbose:
                        print(len(self.enc_debug_arr))
                    raise BusShutdownException

                if channel == 'can':
                    self.slot_can(self.time, data)
                elif channel == 'desired_speed':
                    self.slot_desired_speed(self.time, data)
                else:
                    assert False, channel  # unsupported channel
        except BusShutdownException:
            pass
        assert self.can_errors == 0, self.can_errors  # summary over whole run

    def draw(self):
        """
        Debug - draw encoders
        """
#        draw(self.enc_debug_arr, self.joint_debug_arr)
#        print(self.count_arr)
#        print(self.count_arr[-1])
#        draw_enc_stat(self.count_arr)
#        draw_enc_stat(self.debug_odo)
        draw_4wd( self.speed_debug_arr, self.joint_debug_arr )
#        draw_enc_stat(self.debug_downdrops_front)
#        draw_enc_stat(self.debug_downdrops_rear)

# vim: expandtab sw=4 ts=4



import unittest
import datetime
from unittest.mock import MagicMock, patch, call

from osgar.drivers.lora import LoRa, parse_lora_packet, split_lora_buffer, parse_my_cmd
from osgar.bus import Bus


class LoRaTest(unittest.TestCase):

    def test_parse_packet(self):
        addr, data = parse_lora_packet(b'4|alive\r\n')
        self.assertEqual(addr, [4])
        self.assertEqual(data, b'alive')
        addr, data = parse_lora_packet(b'4|1|2|pose(1, 2, 3)\r\n')
        self.assertEqual(addr, [4, 1, 2])
        self.assertEqual(data, b'pose(1, 2, 3)')

    def test_autodetect(self):
        SAMPLE_DATA = b'1|cmd=home\n'
        logger = MagicMock(write=MagicMock(return_value=datetime.timedelta()))
        bus = Bus(logger)
        c = LoRa(bus=bus.handle('lora'), config={'device_id': 4})
        tester = bus.handle('tester')
        tester.register('raw')
        bus.connect('tester.raw', 'lora.raw')
        c.start()
        tester.publish('raw', SAMPLE_DATA)
        c.request_stop()
        c.join()
        self.assertEqual(c.device_id, 4)

    def test_split_lora_buffer(self):
        self.assertEqual(split_lora_buffer(b'1|ble'), (b'1|ble', b''))
        self.assertEqual(split_lora_buffer(b'1|ble\n3|'), (b'3|', b'1|ble\n'))

    def test_invalid_packet(self):
        addr, data = parse_lora_packet(b'|pose(1, 2, 3)\r\n')
        self.assertIsNone(addr)
        self.assertEqual(data, b'|pose(1, 2, 3)\r\n')

    def test_autodetect_bug(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        bus = Bus(logger)
        c = LoRa(bus=bus.handle('lora'), config={})  # force autodetection
        tester = bus.handle('tester')
        tester.register('raw')
        bus.connect('tester.raw', 'lora.raw')
        c.start()
        tester.publish('raw', b'4|alive\n')
        tester.publish('raw', b'4|alive-97')
        tester.publish('raw', b'21\n')
        c.request_stop()
        c.join()
        self.assertEqual(c.device_id, 4)

    def test_3rd_party_packet(self):
        # there are internal debug messages and also 3rd party packets, which we definetely
        # do not want to re-transmmit and process
        # In particular there was a bug that missing '|' generated empty addr list
        self.assertIsNone(parse_lora_packet(b'dhcps: send_off')[0])

        self.assertIsNone(parse_lora_packet(b'0|data')[0])
        self.assertIsNone(parse_lora_packet(b'7|data')[0])
        self.assertIsNone(parse_lora_packet(b'42|data')[0])

        # validate all addresses in chain
        self.assertIsNone(parse_lora_packet(b'4|7|data')[0])

    def test_parse_cmd(self):
        self.assertEqual(parse_my_cmd(4, b'4:GoHome:14151'), b'GoHome')
        self.assertEqual(parse_my_cmd(4, b'0:Pause:11'), b'Pause')  # 0 for ALL
        self.assertIsNone(parse_my_cmd(2, b'4:GoHome:14151'))
        self.assertIsNone(parse_my_cmd(2, b'nonsense'))

    def test_send_cmd(self):
        logger = MagicMock()
        bus = Bus(logger)
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        c = LoRa(bus=bus.handle('lora'), config={'device_id':3})
        tester = bus.handle('tester')
        tester.register('raw')
        bus.connect('lora.cmd', 'tester.cmd')
        bus.connect('tester.raw', 'lora.raw')
        c.start()
        tester.publish('raw', b'1|3:GoHome:1234\n')
        c.request_stop()
        c.join()
        self.assertEqual(tester.listen()[2], b'GoHome')

    def test_control_center(self):
        logger = MagicMock()
        bus = Bus(logger)
        logger.write = MagicMock(return_value=datetime.timedelta(seconds=97))
        c = LoRa(bus=bus.handle('lora'), config={'device_id':1})
        app = bus.handle('app')
        app.register('cmd')
        tester = bus.handle('serial')
        tester.register('raw')
        bus.connect('app.cmd', 'lora.cmd')
        bus.connect('lora.raw', 'serial.raw')
        c.start()
        app.publish('cmd', [3, b'Pause'])
        c.request_stop()
        c.join()
        self.assertEqual(tester.listen()[2], b'alive\n')
        self.assertEqual(tester.listen()[2], b'3:Pause:97\n')

    def test_report_artifact(self):
        bus = MagicMock()
        c = LoRa(bus=bus, config={'device_id':1})
        with patch('osgar.node.Node.update') as p:
            p.return_value = 'raw'
            c.raw = b"3|['TYPE_BACKPACK', 3506, -18369, -752]\n"
            c.update()
            self.assertEqual(bus.method_calls,
                             [call.register('raw', 'cmd', 'robot_status', 'artf', 'artf_xyz'),
                              call.publish('artf', [3, ['TYPE_BACKPACK', 3506, -18369, -752]]),
                              call.publish('raw', b"3|['TYPE_BACKPACK', 3506, -18369, -752]\n")]
                             )

    def test_on_radio(self):
        # virtual world
        data = [b'A0F150L', b"['TYPE_RESCUE_RANDY', 15982, 104845, 3080]\n"]
        bus = MagicMock()
        dev = LoRa(bus=bus, config={'device_id':1})
        bus.reset_mock()
        dev.on_radio(data)
        bus.publish.assert_called()

        data = [b'A0F150L', b'[11896, 7018, -11886]\n']
        bus.reset_mock()
        dev.on_radio(data)
        bus.publish.assert_not_called()

# vim: expandtab sw=4 ts=4
